
import os

import numpy as np

from .catalogueconstructor import CatalogueConstructor
from .matplotlibplot import plot_centroids, plot_waveforms_histogram, plot_isi
from .tools import get_neighborhood



_summary_catalogue_clusters_template = """
Cluster {label}
Peak channel (abs): {extremum_channel_abs}
Peak channel (local to group): {extremum_channel}
Peak amplitude MAD: {extremum_amplitude}
Peak amplitude (µV): {extremum_amplitude_uV}

"""
def summary_catalogue_clusters(dataio, chan_grp=None, labels=None, label=None,
        show_channels=None, neighborhood_radius=None):
    
    
    cc =CatalogueConstructor(dataio, chan_grp=chan_grp)
    
    if labels is None and label is None:
        labels = cc.positive_cluster_labels
    if label is not None:
        labels = [label]
    
    # channel index absolut to datasource
    channel_abs = dataio.channel_groups[chan_grp]['channels']

    if show_channels is None:
        if neighborhood_radius is None:
            if len(channel_abs)>10:
                show_channels = False
            else:
                show_channels = True
        else:
            show_channels = True
    
    for l, label in enumerate(labels):

        ind = cc.index_of_label(label)
        cluster = cc.clusters[ind]

        extremum_channel = cluster['extremum_channel']
        
        if extremum_channel>=0:
            extremum_channel_abs = channel_abs[extremum_channel]
        else:
            extremum_channel = None
            extremum_channel_abs = None
        
        extremum_amplitude = cluster['extremum_amplitude']
        extremum_amplitude_uV = np.nan
        if cc.dataio.datasource.bit_to_microVolt is not None and extremum_channel is not None:
            extremum_amplitude_uV = extremum_amplitude * cc.signals_mads[extremum_channel] * cc.dataio.datasource.bit_to_microVolt
            
            
        
        d = dict(label=label,
                    extremum_channel=extremum_channel,
                    extremum_channel_abs=extremum_channel_abs,
                    extremum_amplitude=extremum_amplitude,
                    extremum_amplitude_uV=extremum_amplitude_uV,
                    )
        text = _summary_catalogue_clusters_template.format(**d)
        
        print(text)
        
        import matplotlib.pyplot as plt
        fig, axs = plt.subplots(ncols=2)
        plot_centroids(cc, labels=[label,], ax=axs[0], show_channels=show_channels, neighborhood_radius=neighborhood_radius)
        axs[0].set_title('cluster {}'.format(label))
        
        if dataio.datasource.bit_to_microVolt is None:
            units = 'MAD'
            title = 'Amplitude in MAD (STD) ratio'
        else:
            units = 'uV'
            title = 'Amplitude μV'
        
        plot_waveforms_histogram(cc, label=label, ax=axs[1], channels=[extremum_channel], units=units)
        axs[1].set_title('Hist on channel {}'.format(extremum_channel_abs))
        axs[1].set_ylabel(title)



_summary_noise_template = """
Noise mesured with mesured with MAD (=robust STD)
  Channel group {chan_grp}
  Nb_channel:  {nb_chan}
  Noise range: {min_mad:.2f} - {max_mad:.2f} μV
  By channel noise: {reduce_noise_vector} μV
  Average noise along channel: {noise_mean}  μV
  Threshold:  {relative_threshold} *MAD = {thresh_uV} μV
"""

def summary_noise(dataio, chan_grp=None):
    cc =CatalogueConstructor(dataio, chan_grp=chan_grp)
    
    channels = dataio.channel_groups[chan_grp]['channels']
    
    assert cc.dataio.datasource.bit_to_microVolt is not None, 'bit_to_microVolt is not provided for this data source'
    
    noise_uV = cc.signals_mads * cc.dataio.datasource.bit_to_microVolt
    
    noise_txt = [ '{:.1f}'.format(e) for e in noise_uV]
    
    if len(channels)>8:
        reduce_noise_vector = "[{} ... {}]".format(' '.join(noise_txt[:4]),' '.join(noise_txt[-4:]))
    else:
        reduce_noise_vector = "[{}]".format(' '.join(noise_txt))
    
    
    relative_threshold = cc.info['peak_detector']['relative_threshold']
    
    d =dict(chan_grp=chan_grp,
        nb_chan=len(channels),
        min_mad=min(noise_uV),
        max_mad=max(noise_uV),
        reduce_noise_vector=reduce_noise_vector,
        noise_mean=np.mean(noise_uV),
        relative_threshold=relative_threshold,
        thresh_uV=relative_threshold * np.mean(noise_uV),
        
    )
    
    text = _summary_noise_template.format(**d)
    
    print(text)




_summary_after_peeler_clusters_template = """
Cluster {label}
Max on channel (abs): {extremum_channel_abs}
Max on channel (local to group): {extremum_channel}
Peak amplitude MAD: {extremum_amplitude:.1f}
Peak amplitude (µV): {extremum_amplitude_uV:.1f}
Nb spikes : {nb_spike}

"""

def summary_after_peeler_clusters(dataio, catalogue=None, chan_grp=None, labels=None, label=None,
            show_channels=None, neighborhood_radius=None):
    if chan_grp is None:
        chan_grp = min(dataio.channel_groups.keys())
    
    
    #~ cc =CatalogueConstructor(dataio, chan_grp=chan_grp)
    if catalogue is None:
        catalogue = dataio.load_catalogue(chan_grp=chan_grp)
    
    clusters = catalogue['clusters']
    
    if labels is None and label is None:
        labels = clusters['cluster_label']
    if label is not None:
        labels = [label]
    
    channel_abs = dataio.channel_groups[chan_grp]['channels']
 
    if show_channels is None:
        if neighborhood_radius is None:
            if len(channel_abs)>10:
                show_channels = False
            else:
                show_channels = True
        else:
            show_channels = True
    
    figs = []
    for l, label in enumerate(labels):
        ind = np.nonzero(clusters['cluster_label']==label)[0][0]
        
        cluster = clusters[ind]
        
        extremum_channel = cluster['extremum_channel']
        if extremum_channel>=0:
            extremum_channel_abs = channel_abs[extremum_channel]
        else:
            extremum_channel = None
            extremum_channel_abs = None

        extremum_amplitude = cluster['extremum_amplitude']
        extremum_amplitude_uV = np.nan
        if dataio.datasource.bit_to_microVolt is not None and extremum_channel is not None:
            extremum_amplitude_uV = extremum_amplitude * catalogue['signals_mads'][extremum_channel] * dataio.datasource.bit_to_microVolt

        
        nb_spike = 0
        for seg_num in range(dataio.nb_segment):
            all_spikes = dataio.get_spikes(seg_num=seg_num, chan_grp=chan_grp)
            nb_spike += np.sum(all_spikes['cluster_label'] == label)
            

        import matplotlib.pyplot as plt
        fig, axs = plt.subplots(ncols=2, nrows=2)
        figs.append(fig)
        axs[0, 0].remove()
        # centroids
        ax = axs[0, 1]
        plot_centroids(catalogue, dataio=dataio, labels=[label,], ax=ax, show_channels=show_channels, neighborhood_radius=neighborhood_radius)
        ax.set_title('cluster {}'.format(label))
        
        
        # waveform density
        ax = axs[1, 1]
        if dataio.datasource.bit_to_microVolt is None:
            units = 'MAD'
            title = 'Amplitude in MAD (STD) ratio'
        else:
            units = 'uV'
            title = 'Amplitude μV'
        
            
        if neighborhood_radius is None:
            plot_channels_hist = [extremum_channel]
        else:
            neihb = get_neighborhood(dataio.get_geometry(chan_grp=chan_grp) , neighborhood_radius)
            plot_channels_hist,  = np.nonzero(neihb[extremum_channel])    # local index (to group)
            
        plot_waveforms_histogram(catalogue, dataio=dataio, label=label, ax=ax, channels=plot_channels_hist,
                        units=units)
        ax.set_title(title)
        
        # ISI
        ax = axs[1, 0]
        plot_isi(dataio, catalogue=catalogue, chan_grp=chan_grp, label=label, ax=ax, bin_min=0, bin_max=100, bin_size=1.)
        ax.set_title('ISI (ms)')

        d =dict(chan_grp=chan_grp,
                    label=label,
                    extremum_channel=extremum_channel,
                    extremum_channel_abs=extremum_channel_abs,
                    extremum_amplitude=extremum_amplitude,
                    extremum_amplitude_uV=extremum_amplitude_uV,
                    nb_spike=nb_spike,
            
        )
        
        text = _summary_after_peeler_clusters_template.format(**d)
        
        #~ print(text)
        
        ax.figure.text(.05, .75, text,  va='center') #, ha='center')
    
    return figs


def generate_report(dataio, export_path=None, neighborhood_radius=None) :
    if export_path is None:
        export_path = os.path.join(dataio.dirname, 'report')
    
    
    for chan_grp in dataio.channel_groups.keys():
        if not dataio.is_spike_computed(chan_grp=chan_grp):
            continue
        
        catalogue = dataio.load_catalogue(chan_grp=chan_grp)
        
        if catalogue is None:
            continue
        
        clusters = catalogue['clusters']
        
        # TODO with cell_label
        labels = clusters['cluster_label']
        
        path = os.path.join(export_path, 'chan_grp {}'.format(chan_grp), 'figures')
        if not os.path.exists(path):
            os.makedirs(path)
        
        for label in labels:
            figs = summary_after_peeler_clusters(dataio, chan_grp=chan_grp, label=label, neighborhood_radius=neighborhood_radius)
            figs[0].savefig(os.path.join(path, 'summary_cluster_{}.png'.format(label)))
    


from tridesclous import *

import matplotlib.pyplot as plt

from tridesclous.tests.testingtools import setup_catalogue

from tridesclous.tests.testingtools import ON_CI_CLOUD
import pytest

import shutil
import os
import time


def setup_module():
    setup_catalogue('test_metrics', dataset_name='olfactory_bulb')

def teardown_module():
    if not(os.environ.get('APPVEYOR') in ('true', 'True')):
        # this fix appveyor teardown_module bug
        shutil.rmtree('test_metrics')
    
    

def test_all_metrics():
    dataio = DataIO(dirname='test_metrics')
    cc = CatalogueConstructor(dataio=dataio)
    
    #~ cc.compute_spike_waveforms_similarity()
    
    t0 = time.perf_counter()
    cc.compute_cluster_similarity()
    t1 = time.perf_counter()
    print('compute_cluster_similarity', t1-t0)
    print(cc.cluster_similarity)
    
    t0 = time.perf_counter()
    cc.compute_cluster_ratio_similarity()
    t1 = time.perf_counter()
    print('compute_cluster_ratio_similarity', t1-t0)
    print(cc.cluster_ratio_similarity)
    
    t0 = time.perf_counter()
    cc.compute_spike_silhouette()
    t1 = time.perf_counter()
    print('compute_spike_silhouette', t1-t0)
    print(cc.spike_silhouette)

    


@pytest.mark.skipif(ON_CI_CLOUD, reason='ON_CI_CLOUD')
def test_cluster_ratio():
    dataio = DataIO(dirname='test_metrics')
    cc = CatalogueConstructor(dataio=dataio)
    
    cc.compute_cluster_similarity()
    cc.compute_cluster_ratio_similarity()

    for name in('cluster_similarity', 'cluster_ratio_similarity'):
        d = getattr(cc, name).copy()
        fig, ax = plt.subplots()
        im  = ax.matshow(d, cmap='viridis')
        im.set_clim(0,1)
        fig.colorbar(im)
        ax.set_title(name)
    
    


if __name__ == '__main__':
    setup_module()
    
    test_all_metrics()
    test_cluster_ratio()
    
    #~ plt.show()


from ..peeler import Peeler


from pyqtgraph.util.mutex import Mutex

import pyacq
from pyacq import Node, register_node_type, ThreadPollInput

from ..peeler import _dtype_spike




class PeelerThread(ThreadPollInput):
    def __init__(self, input_stream, output_streams, peeler,in_group_channels, geometry,
                        timeout = 200, parent = None):
        
        ThreadPollInput.__init__(self, input_stream,  timeout=timeout, return_data=True, parent = parent)
        self.output_streams = output_streams
        self.peeler = peeler
        self.in_group_channels = in_group_channels
        self.geometry = geometry
        
        self.sample_rate = input_stream.params['sample_rate']
        self.total_channel = self.input_stream().params['shape'][1]
        
        self.mutex = Mutex()
    
    def process_data(self, pos, sigs_chunk):
        #TODO maybe remove this
        #~ print('process_data', sigs_chunk.shape[0], self.peeler.chunksize)
        assert sigs_chunk.shape[0] == self.peeler.chunksize, 'PeelerThread chunksize is BAD!! {} {}'.format(sigs_chunk.shape[0], self.peeler.chunksize)
        
        with self.mutex:
            #take only channels concerned
            sigs_chunk = sigs_chunk[:, self.in_group_channels]
            #~ print('pos', pos)
            #~ print(sigs_chunk.shape, sigs_chunk.dtype)
            #~ print('signals_medians', self.peeler.signalpreprocessor.signals_medians)
            sig_index, preprocessed_chunk, total_spike, spikes  = self.peeler.process_one_chunk(pos, sigs_chunk)
            #~ print('sig_index', sig_index)
            
            #~ print('total_spike', total_spike, len(spikes))
            #~ print('sig_index', sig_index, preprocessed_chunk.shape)
            
            self.output_streams['signals'].send(preprocessed_chunk, index=sig_index)
            #~ if spikes is not None and spikes.size>0:
            if spikes.size>0:
                self.output_streams['spikes'].send(spikes, index=total_spike)
        
    
    def change_params(self, **peeler_params):
        #~ print('PeelerThread.change_params')
        with self.mutex:
            # complicated hack pad_width can be changed on the fly
            # because it break the index stream with advance or delay
            # so we always forec the prvious one!!!!
            if hasattr(self.peeler, 'peeler_engine') and hasattr(self.peeler.peeler_engine, 'signalpreprocessor'):
                #~ print(self.peeler.peeler_engine.signalpreprocessor.pad_width)
                prev_pad_width = self.peeler.peeler_engine.signalpreprocessor.pad_width
                #~ if 'pad_width' in catalogue['signal_preprocessor_params']:
                peeler_params['catalogue']['signal_preprocessor_params']['pad_width'] = prev_pad_width
                #~ print('*'*50)
                #~ print('Force prev_pad_width', prev_pad_width)
                #~ print('*'*50)
            else:
                prev_pad_width = None

            self.peeler.change_params(**peeler_params)
            
            buffer_spike_index = self.output_streams['spikes'].last_index
            #~ print('buffer_spike_index', buffer_spike_index)
            
            # TODO check tha pad_width have not changed bechause
            # head index will be out
            
            
            
            self.peeler.initialize_online_loop(sample_rate=self.sample_rate,
                                                nb_channel=len(self.in_group_channels),
                                                source_dtype=self.input_stream().params['dtype'],
                                                geometry=self.geometry,
                                                )
            
            #~ print('*'*50)
            #~ print(self.peeler.peeler_engine.signalpreprocessor.pad_width)
            #~ print('*'*50)
            
            #~ print('self.peeler.peeler_engine.total_spike', self.peeler.peeler_engine.total_spike)
            self.peeler.peeler_engine.total_spike = buffer_spike_index
            #~ print('self.peeler.peeler_engine.total_spike', self.peeler.peeler_engine.total_spike)
            
            
            

class OnlinePeeler(Node):
    """
    Wrapper on top of Peeler class to make a pyacq Node.
    And so to have on line spike sorting!!
    """
    _input_specs = {'signals' : dict(streamtype = 'signals')}
    _output_specs = {'signals' : dict(streamtype = 'signals'),
                                'spikes': dict(streamtype='events', shape = (-1, ),  dtype=_dtype_spike),
                                }

    def __init__(self , **kargs):
        Node.__init__(self, **kargs)
    
    def _configure(self, in_group_channels=None, catalogue=None, chunksize=None,
                                    internal_dtype='float32', 
                                    geometry=None,
                                    **peeler_params):
        
        #~ if 'engine' in peeler_engine_kargs:
            #~ peeler_engine = peeler_engine_kargs.pop('engine')
        #~ print('peeler_engine', peeler_engine)
        
        assert geometry is not None
        
        self.in_group_channels = in_group_channels
        self.catalogue = catalogue
        self.chunksize = chunksize
        self.internal_dtype = internal_dtype
        
        self.geometry = geometry
        self.peeler_params = peeler_params
        
        

    def after_input_connect(self, inputname):
        self.total_channel = self.input.params['shape'][1]
        self.sample_rate = self.input.params['sample_rate']
        
        # internal dtype (for waveforms) will also be the output dtype
        self.outputs['signals'].spec['dtype'] = self.internal_dtype
        self.outputs['signals'].spec['shape'] = (-1, len(self.in_group_channels))
        self.outputs['signals'].spec['sample_rate'] = self.input.params['sample_rate']
    
    def after_output_configure(self, inputname):
        channel_info = self.input.params.get('channel_info', None)
        if channel_info is not None:
            channel_info = [channel_info[c] for c in self.in_group_channels]
            self.outputs['signals'].params['channel_info'] = channel_info
    
    def _initialize(self):
        
        self.peeler = Peeler(dataio=None)
        #~ self.peeler.change_params(catalogue=self.catalogue, 
                                        #~ chunksize=self.chunksize, internal_dtype=self.internal_dtype,)
        
        self.thread = PeelerThread(self.input, self.outputs, self.peeler, self.in_group_channels, self.geometry)
        self.change_catalogue(self.catalogue)
        
    def _start(self):
        #~ self.peeler.initialize_online_loop(sample_rate=self.input.params['sample_rate'],
                                            #~ nb_channel=len(self.in_group_channels),
                                            #~ source_dtype=self.input.params['dtype'])
        self.thread.start()
        
    def _stop(self):
        self.thread.stop()
        self.thread.wait()

        
    def _close(self):
        pass
    
    def change_catalogue(self, catalogue):
        print('change_catalogue', )
        #~ print(catalogue['label_to_index'])
        print('self.peeler_params', self.peeler_params)
        self.catalogue = catalogue
        #~ self.thre.change_params(catalogue=self.catalogue, 
                                        #~ chunksize=self.chunksize, internal_dtype=self.internal_dtype,)
        self.thread.change_params(catalogue=catalogue, 
                                        chunksize=self.chunksize, internal_dtype=self.internal_dtype,
                                        **self.peeler_params)
                                        
                                        
        
    
    
#~ register_node_type(OnlinePeeler)



"""

.. autoclass:: Peeler
   :members:

"""

import os
import json
from collections import OrderedDict, namedtuple
import time

import numpy as np
import scipy.signal


from .peeler_tools import _dtype_spike


from tqdm import tqdm

#~ from .peeler_engine_testing import PeelerEngineTesting
from .peeler_engine_geometry import PeelerEngineGeometrical
from .peeler_engine_geometry_cl import PeelerEngineGeometricalCl




peeler_engines = {
    #~ 'testing' : PeelerEngineTesting,
    'geometrical' : PeelerEngineGeometrical,
    'geometrical_opencl' : PeelerEngineGeometricalCl,
}



class Peeler:
    """
    The peeler is core of spike sorting itself.
    It basically do a *template matching* on a signals.
    
    This class nedd a *catalogue* constructed by :class:`CatalogueConstructor`.
    Then the compting is applied chunk chunk on the raw signal itself.
    
    So this class is the same for both offline/online computing.
    
    At each chunk, the algo is basically this one:
      1. apply the processing chain (filter, normamlize, ....)
      2. Detect peaks
      3. Try to classify peak and detect the *jitter*
      4. With labeled peak create a prediction for the chunk
      5. Substract the prediction from the processed signals.
      6. Go back to **2** until there is no peak or only peaks that can't be labeled.
      7. return labeld spikes from this or previous chunk and the processed signals (for display or recoding)
    
    The main difficulty in the implemtation is to deal with edge because spikes 
    waveforms can spread out in between 2 chunk.
    
    Note that the global latency depend on this paramters:
      * pad_width
      * chunksize

    
    """
    def __init__(self, dataio):
        #for online dataio is None
        self.dataio = dataio

    def __repr__(self):
        t = "Peeler <id: {}> \n  workdir: {}\n".format(id(self), self.dataio.dirname)
        
        return t

    def change_params(self, catalogue=None, engine='geometrical', internal_dtype='float32',
                        chunksize=1024, speed_test_mode=False, **params):
        """
        
        speed_test_mode: only for offline mode create a log file with 
                run time for each buffers
        """
        assert catalogue is not None
        
        self.catalogue = catalogue
        self.engine_name = engine
        self.internal_dtype = internal_dtype
        self.chunksize = chunksize
        self.speed_test_mode = speed_test_mode
        
        self.peeler_engine = peeler_engines[engine]()
        self.peeler_engine.change_params(catalogue=catalogue, internal_dtype=internal_dtype, chunksize=chunksize, **params)
    
    def process_one_chunk(self,  pos, sigs_chunk):
        # this is for online
        return self.peeler_engine.process_one_chunk(pos, sigs_chunk)
        
        #~ abs_head_index, preprocessed_chunk, self.total_spike, all_spikes,  = self.peeler_engine.process_one_chunk(pos, sigs_chunk)
        #~ print(pos, sigs_chunk.shape, abs_head_index, preprocessed_chunk.shape)
        #~ return abs_head_index, preprocessed_chunk, self.total_spike, all_spikes
    
    def initialize_online_loop(self, sample_rate=None, nb_channel=None, source_dtype=None, geometry=None):
        # global initialize
        self.peeler_engine.initialize(sample_rate=sample_rate, nb_channel=nb_channel,
                        source_dtype=source_dtype, already_processed=False, geometry=geometry)
        self.peeler_engine.initialize_before_each_segment(already_processed=False)
    
    def run_offline_loop_one_segment(self, seg_num=0, duration=None, progressbar=True):
        chan_grp = self.catalogue['chan_grp']

        if duration is not None:
            length = int(duration*self.dataio.sample_rate)
        else:
            length = self.dataio.get_segment_length(seg_num)
        
        # check if the desired length is already computed or not for this particular segment
        already_processed = self.dataio.already_processed(seg_num=seg_num, chan_grp=chan_grp, length=length)
        
        self.peeler_engine.initialize_before_each_segment(already_processed=already_processed)
        #~ print('run_offline_loop_one_segment already_processed', already_processed)
        
        if already_processed:
            # ready from "processed'
            signal_type = 'processed'
        else:
            # read from "initial" 
            # activate signal processor
            signal_type = 'initial'
        
            #initialize engines
            self.dataio.reset_processed_signals(seg_num=seg_num, chan_grp=chan_grp, dtype=self.internal_dtype)
        
        self.dataio.reset_spikes(seg_num=seg_num, chan_grp=chan_grp, dtype=_dtype_spike)

        iterator = self.dataio.iter_over_chunk(seg_num=seg_num, chan_grp=chan_grp, chunksize=self.chunksize, 
                                                    i_stop=length, signal_type=signal_type)
        if progressbar:
            iterator = tqdm(iterable=iterator, total=length//self.chunksize)
        
        if self.speed_test_mode:
            process_run_times = []
        
        for pos, sigs_chunk in iterator:
            if self.speed_test_mode:
                t0 = time.perf_counter()
            
            sig_index, preprocessed_chunk, total_spike, spikes = self.peeler_engine.process_one_chunk(pos, sigs_chunk)
            
            if self.speed_test_mode:
                t1 = time.perf_counter()
                process_run_times.append(t1-t0)
                

            
            
            if sig_index<=0:
                continue
            
            if not already_processed:
                # save preprocessed_chunk to file
                self.dataio.set_signals_chunk(preprocessed_chunk, seg_num=seg_num,chan_grp=chan_grp,
                            i_start=sig_index-preprocessed_chunk.shape[0], i_stop=sig_index,
                            signal_type='processed')
            
            if spikes is not None and spikes.size>0:
                self.dataio.append_spikes(seg_num=seg_num, chan_grp=chan_grp, spikes=spikes)
        
        extra_spikes = self.peeler_engine.get_remaining_spikes()
        if extra_spikes is not None:
            
            if extra_spikes.size>0:
                self.dataio.append_spikes(seg_num=seg_num, chan_grp=chan_grp, spikes=extra_spikes)
        
        if not already_processed:
            self.dataio.flush_processed_signals(seg_num=seg_num, chan_grp=chan_grp, processed_length=int(sig_index))
            
        self.dataio.flush_spikes(seg_num=seg_num, chan_grp=chan_grp)
        
        if self.speed_test_mode:
            process_run_times = np.array(process_run_times, dtype='float64')
            log_path = self.dataio.get_log_path(chan_grp=chan_grp)
            filename = os.path.join(log_path, 'peeler_run_times_seg{}.npy'.format(seg_num))
            np.save(filename, process_run_times)

    def run(self, duration=None, progressbar=True):
        assert hasattr(self, 'catalogue'), 'So peeler.change_params first'
        
        chan_grp = self.catalogue['chan_grp']
        
        duration_per_segment = self.dataio.get_duration_per_segments(duration)
        
        already_processed_segs = []
        for seg_num in range(self.dataio.nb_segment):
            length = int(duration_per_segment[seg_num]*self.dataio.sample_rate)
            
            # check if the desired length is already computed or not
            already_processed = self.dataio.already_processed(seg_num=seg_num, chan_grp=chan_grp, length=length)
            already_processed_segs.append(already_processed)
        
        kargs = {}
        kargs['sample_rate'] = self.dataio.sample_rate
        kargs['nb_channel'] = self.dataio.nb_channel(chan_grp)
        if any(already_processed_segs):
            kargs['source_dtype'] = self.internal_dtype
        else:
            kargs['source_dtype'] = self.dataio.source_dtype
        kargs['geometry'] = self.dataio.get_geometry(chan_grp)
        kargs['already_processed'] =  all(already_processed_segs)
        self.peeler_engine.initialize(**kargs)
        
        for seg_num in range(self.dataio.nb_segment):
            self.run_offline_loop_one_segment(seg_num=seg_num, duration=duration_per_segment[seg_num], progressbar=progressbar)
    
    # old alias just in case
    run_offline_all_segment = run
    
    
    def get_run_times(self, chan_grp=0, seg_num=0):
        """
        need speed_test_mode=True in params
        """
        p = self.dataio.get_log_path(chan_grp=chan_grp)
        filename = os.path.join(p, 'peeler_run_times_seg{}.npy'.format(seg_num))
        run_times = np.load(filename)
        return run_times
    



import time
import copy
from pprint import pprint

import numpy as np

from .cltools import HAVE_PYOPENCL
from .cluster import HAVE_ISOSPLIT5

try:
    import numba
    HAVE_NUMBA = True
except ImportError:
    HAVE_NUMBA = False


# nb of channel to become sparse
limit_dense_sparse = 4


_default_catalogue_params = {
    'duration': 300.,
    
    'chunksize': 1024,
    'mode': 'dense', # 'sparse'
    'sparse_threshold': None, # 1.5
    'memory_mode': 'memmap',
    'n_spike_for_centroid':350,
    'n_jobs' :-1,

    'preprocessor': {
        'highpass_freq': 300.,
        'lowpass_freq': 5000.,
        'smooth_size': 0,
        
        'pad_width': -1,   # this auto
        'engine': 'numpy',
        'common_ref_removal':False,
    },
    'peak_detector': {
        'method' : 'global',
        'engine': 'numpy',
        'peak_sign': '-',
        'relative_threshold': 5.,
        'peak_span_ms': .7,
        'adjacency_radius_um' : None,
        'smooth_radius_um' : None,
        
    },
    'noise_snippet': {
        'nb_snippet': 300,
    },
    'extract_waveforms': {
        'wf_left_ms': -1.,
        'wf_right_ms': 1.5,
        'wf_left_long_ms': -2.5,
        'wf_right_long_ms': 3.5,
    },
    'clean_peaks': {
        #~ 'alien_value_threshold': None,
        'alien_value_threshold':-1., # equivalent to None (normally)
        #~ 'alien_value_threshold': np.nan,
        'mode': 'extremum_amplitude',
    },
    'peak_sampler': {
        'mode': 'rand',
        'nb_max': 20000,
        #~ 'nb_max_by_channel': None,
        'nb_max_by_channel': 600.,
    },
    #~ 'clean_waveforms': {
        #~ 'alien_value_threshold': None,
    #~ },
    'feature_method': 'pca_by_channel',
    'feature_kargs': {},
    'cluster_method': 'pruningshears',
    'cluster_kargs': {},
    
    
    'clean_cluster' :{
        'apply_auto_split': True,
        'apply_trash_not_aligned': True,
        'apply_auto_merge_cluster': True,
        'apply_trash_low_extremum': True,
        'apply_trash_small_cluster': True,
    },
    
    
    'make_catalogue':{
        'inter_sample_oversampling':False,
        'subsample_ratio': 'auto',
        'sparse_thresh_level2': 1.5,
        'sparse_thresh_level2': 3,
    }
}





def get_auto_params_for_catalogue(dataio=None, chan_grp=0,
                            nb_chan=None, sample_rate=None,
                            context='offline'):
    """
    Automatic selection of parameters.
    This totally empiric paramerters.
    """
    
    assert context in ('offline', 'online')
    
    params = copy.deepcopy(_default_catalogue_params)

    # TODO make this more complicated
    #  * by detecting if dense array or not.
    #  * better method sleection


    if dataio is None:
        # case online no DataIO
        assert nb_chan is not None
        assert sample_rate is not None
        # use less because chunksize is set outsied
        params['chunksize'] = int(sample_rate * 0.05)
        params['duration'] = 10. # debug
    else:
        # case offline with a DataIO
        nb_chan = dataio.nb_channel(chan_grp=chan_grp)
        sample_rate = dataio.sample_rate
        total_duration = sum(dataio.get_segment_length(seg_num=seg_num) / dataio.sample_rate for seg_num in range(dataio.nb_segment))
        # auto chunsize of 100 ms
        params['chunksize'] = int(sample_rate * 1.0)
        params['duration'] = 601.
        
        # segment durartion is not so big then take the whole duration
        # to avoid double preprocessing (catalogue+peeler)
        if params['duration'] * 2 > total_duration:
            params['duration'] = total_duration
    
    #~ if nb_chan == 1:
        #~ params['mode'] = 'dense'
        #~ params['adjacency_radius_um'] = 0.
        #~ params['sparse_threshold'] = 1.5
        
        #~ params['peak_detector']['method'] = 'global'
        #~ params['peak_detector']['engine'] = 'numpy'
        #~ params['peak_detector']['smooth_radius_um' ] = None


        #~ params['peak_sampler']['mode'] = 'rand'
        #~ params['peak_sampler']['nb_max'] = 20000
        
        #~ params['feature_method'] = 'global_pca'
        #~ params['feature_kargs'] = {'n_components' : 4 }
        
        #~ params['cluster_method'] = 'dbscan_with_noise'
        #~ params['cluster_kargs'] = {}
        

        
        #~ params['clean_cluster_kargs'] = {'too_small' : 20 }        
    
    #~ elif nb_chan <=4:
    if nb_chan <= limit_dense_sparse:
    #~ if nb_chan <=8:
    
        params['mode'] = 'dense'
        #~ params['adjacency_radius_um'] = 0.
        params['sparse_threshold'] = 1.5
        
        #~ params['peak_detector']['method'] = 'global'
        #~ params['peak_detector']['engine'] = 'numpy'

        params['peak_detector']['method'] = 'geometrical'
        params['peak_detector']['engine'] = 'numba'
        
        params['peak_detector']['adjacency_radius_um'] = 200. # useless
        params['peak_detector']['smooth_radius_um' ] = None


        params['peak_sampler']['mode'] = 'rand'
        params['peak_sampler']['nb_max'] = 20000
        
        params['feature_method'] = 'global_pca'
        if nb_chan in (1,2):
            n_components = 5
        else:
            n_components = int(nb_chan*2)
        params['feature_kargs'] = {'n_components' : n_components }
        
        
        params['cluster_method'] = 'pruningshears'
        params['cluster_kargs']['max_loop'] = max(1000, nb_chan * 10)
        params['cluster_kargs']['min_cluster_size'] = 20
        params['cluster_kargs']['adjacency_radius_um'] = 0.
        params['cluster_kargs']['high_adjacency_radius_um'] = 0.
        
        # necessary for peeler classic
        #~ params['make_catalogue']['inter_sample_oversampling'] = True


        

    else:
        params['mode'] = 'sparse'
        #~ params['adjacency_radius_um'] = 200.
        params['sparse_threshold'] = 1.5

        if nb_chan >= 32 and HAVE_PYOPENCL:
            params['preprocessor']['engine'] = 'opencl'
        params['peak_detector']['method'] = 'geometrical'
        params['peak_detector']['adjacency_radius_um'] = 200.
        #~ params['peak_detector']['smooth_radius_um' ] = 10
        params['peak_detector']['smooth_radius_um' ] = None
        
        
        if HAVE_PYOPENCL:
            params['peak_detector']['engine'] = 'opencl'
        elif HAVE_NUMBA:
            params['peak_detector']['engine'] = 'numba'
        else:
            print('WARNING : peakdetector will be slow install opencl or numba')
            params['peak_detector']['engine'] = 'numpy'
        
        params['peak_sampler']['mode'] = 'rand_by_channel'
        #~ params['extract_waveforms']['nb_max_by_channel'] = 700
        params['peak_sampler']['nb_max_by_channel'] = 1000
        #~ params['peak_sampler']['nb_max_by_channel'] = 1500
        #~ params['peak_sampler']['nb_max_by_channel'] = 3000
        
        
        params['feature_method'] = 'pca_by_channel'
        # TODO change n_components_by_channel depending on channel density
        #~ params['feature_kargs'] = {'n_components_by_channel':5}
        params['feature_kargs'] = {'n_components_by_channel': 3,
                                                        'adjacency_radius_um' :50.,  # this should be greater than cluster 'adjacency_radius_um'
                                                        }
        
        params['cluster_method'] = 'pruningshears'
        params['cluster_kargs']['max_loop'] = max(1000, nb_chan * 20)
        params['cluster_kargs']['min_cluster_size'] = 20
        params['cluster_kargs']['adjacency_radius_um'] = 50.
        params['cluster_kargs']['high_adjacency_radius_um'] = 30.

    if context == 'online':
        params['n_jobs' ] = 1
    
    return params



def get_auto_params_for_peelers(dataio, chan_grp=0):
    nb_chan = dataio.nb_channel(chan_grp=chan_grp)
    params = {}
    
    params['chunksize'] = int(dataio.sample_rate * 1)
    #~ params['chunksize'] = int(dataio.sample_rate * 0.5)
    #~ params['chunksize'] = int(dataio.sample_rate * 0.1)
    #~ params['chunksize'] = int(dataio.sample_rate * 0.033)

    if nb_chan <= limit_dense_sparse:
        params['engine'] = 'geometrical'
        
    else:
        if HAVE_PYOPENCL:
            params['engine'] = 'geometrical_opencl'
        else:
            params['engine'] = 'geometrical'

    
    # DEBUG force 'geometrical'
    params['engine'] = 'geometrical'
        

    return params



from .myqt import QT
import pyqtgraph as pg

import numpy as np

from .base import WidgetBase
from .baselist import ClusterBaseList

from .peelercontroller import spike_visible_modes
from .tools import ParamDialog

from .. import labelcodes

class SpikeModel(QT.QAbstractItemModel):
    def __init__(self, parent =None, controller=None):
        QT.QAbstractItemModel.__init__(self,parent)
        self.controller = controller
        self.refresh_colors()
    
    def columnCount(self , parentIndex):
        return 6
    
    def rowCount(self, parentIndex):
        #~ if not parentIndex.isValid() and self.cc.peak_label is not None:
        if not parentIndex.isValid():
            self.visible_ind, = np.nonzero(self.controller.spikes['visible'])
            return self.visible_ind.size
            
        else :
            return 0
    
    def index(self, row, column, parentIndex):
        if not parentIndex.isValid():
            if column==0:
                childItem = row
            return self.createIndex(row, column, None)
        else:
            return QT.QModelIndex()
    
    def parent(self, index):
        return QT.QModelIndex()
    
    def data(self, index, role):
        if not index.isValid():
            return None
        
        if role not in (QT.Qt.DisplayRole, QT.Qt.DecorationRole):
            return
        
        col = index.column()
        row = index.row()
        
        #~ t_start = 0.
        
        abs_ind = self.visible_ind[row]
        spike = self.controller.spikes[abs_ind]
        
        if np.isnan(spike['jitter']):
            spike_time = (spike['index'])/self.controller.dataio.sample_rate
        else:
            spike_time = (spike['index']+ spike['jitter'])/self.controller.dataio.sample_rate
        
        if role ==QT.Qt.DisplayRole :
            if col == 0:
                return '{}'.format(abs_ind)
            elif col == 1:
                return '{}'.format(spike['segment'])
            elif col == 2:
                return '{}'.format(spike['index'])
            elif col == 3:
                return '{:.2f}'.format(spike['jitter'])
            elif col == 4:
                return '{:.4f}'.format(spike_time)
            elif col == 5:
                return '{}'.format(spike['cluster_label'])
            else:
                return None
        elif role == QT.Qt.DecorationRole :
            if col != 0: return None
            if spike['cluster_label'] in self.icons:
                return self.icons[spike['cluster_label']]
            else:
                return None
        else :
            return None
    
    def flags(self, index):
        if not index.isValid():
            return QT.Qt.NoItemFlags
        return QT.Qt.ItemIsEnabled | QT.Qt.ItemIsSelectable #| Qt.ItemIsDragEnabled

    def headerData(self, section, orientation, role):
        if orientation == QT.Qt.Horizontal and role == QT.Qt.DisplayRole:
            return  ['num', 'seg_num', 'index', 'jitter', 'time', 'cluster_label'][section]
        return
    
    def refresh_colors(self):
        self.icons = { }
        for k, color in self.controller.qcolors.items():
            pix = QT.QPixmap(10,10 )
            pix.fill(color)
            self.icons[k] = QT.QIcon(pix)
        #~ self.icons[-1] = QIcon(':/user-trash.png')
        #~ self.layoutChanged.emit()
        self.refresh()
    
    def refresh(self):
        self.layoutChanged.emit()


class SpikeList(WidgetBase):
    def __init__(self,controller=None, parent=None):
        WidgetBase.__init__(self, parent=parent, controller=controller)
        self.controller = controller
        
        self.layout = QT.QVBoxLayout()
        self.setLayout(self.layout)
        
        self.layout.addWidget(QT.QLabel('<b>All spikes</b>') )
        
        self.combo = QT.QComboBox()
        self.layout.addWidget(self.combo)
        self.combo.addItems(spike_visible_modes)
        self.combo.currentTextChanged.connect(self.change_visible_mode)
        
        self.tree = QT.QTreeView(minimumWidth = 100, uniformRowHeights = True,
                    selectionMode= QT.QAbstractItemView.ExtendedSelection, selectionBehavior = QT.QTreeView.SelectRows,
                    contextMenuPolicy = QT.Qt.CustomContextMenu,)
        
        self.layout.addWidget(self.tree)
        #~ self.tree.customContextMenuRequested.connect(self.open_context_menu)
        
        self.model = SpikeModel(controller=self.controller)
        self.tree.setModel(self.model)
        self.tree.selectionModel().selectionChanged.connect(self.on_tree_selection)

        for i in range(self.model.columnCount(None)):
            self.tree.resizeColumnToContents(i)
        self.tree.setColumnWidth(0,80)
    
    def refresh(self):
        self.model.refresh_colors()
    
    def on_tree_selection(self):
        self.controller.spikes['selected'][:] = False
        for index in self.tree.selectedIndexes():
            if index.column() == 0:
                ind = self.model.visible_ind[index.row()]
                self.controller.spikes['selected'][ind] = True
        self.spike_selection_changed.emit()
    
    def on_spike_selection_changed(self):
        self.tree.selectionModel().selectionChanged.disconnect(self.on_tree_selection)
        
        row_selected, = np.nonzero(self.controller.spikes['selected'][self.model.visible_ind])
        
        if row_selected.size>100:#otherwise this is verry slow
            row_selected = row_selected[:10]
        
        # change selection
        self.tree.selectionModel().clearSelection()
        flags = QT.QItemSelectionModel.Select #| QItemSelectionModel.Rows
        itemsSelection = QT.QItemSelection()
        for r in row_selected:
            for c in range(2):
                index = self.tree.model().index(r,c,QT.QModelIndex())
                ir = QT.QItemSelectionRange( index )
                itemsSelection.append(ir)
        self.tree.selectionModel().select(itemsSelection , flags)

        # set selection visible
        if len(row_selected)>=1:
            index = self.tree.model().index(row_selected[0],0,QT.QModelIndex())
            self.tree.scrollTo(index)

        self.tree.selectionModel().selectionChanged.connect(self.on_tree_selection)        

    def change_visible_mode(self, mode):
        self.controller.change_spike_visible_mode(mode)
        self.cluster_visibility_changed.emit()
        self.model.refresh()

    def open_context_menu(self):
        pass


class ClusterSpikeList(ClusterBaseList):
    
    _special_label = [labelcodes.LABEL_UNCLASSIFIED]
    
    def make_menu(self):
        self.menu = QT.QMenu()
        act = self.menu.addAction('Show all')
        act.triggered.connect(self.show_all)
        act = self.menu.addAction('Hide all')
        act.triggered.connect(self.hide_all)
    



import numpy as np
import time
import os
import shutil

from tridesclous import download_dataset
from tridesclous.dataio import DataIO
from tridesclous.catalogueconstructor import CatalogueConstructor
from tridesclous.tools import median_mad

from matplotlib import pyplot as plt

def test_catalogue_constructor():
    if os.path.exists('test_catalogueconstructor'):
        shutil.rmtree('test_catalogueconstructor')
        
    dataio = DataIO(dirname='test_catalogueconstructor')
    localdir, filenames, params = download_dataset(name='olfactory_bulb')
    #~ localdir, filenames, params = download_dataset(name='locust')
    dataio.set_data_source(type='RawData', filenames=filenames, **params)
    
    channels=range(14)
    #~ channels=list(range(4))
    dataio.add_one_channel_group(channels=channels, chan_grp=0)
    
    cc = CatalogueConstructor(dataio=dataio)
    
    for memory_mode in ['ram', 'memmap']:
        for mode in ['dense', 'sparse']:
        
            print('*'*5)
            print('memory_mode', memory_mode, 'mode', mode)

            if mode == 'dense':
                peak_engine = 'numpy'
                peak_method = 'global'
                adjacency_radius_um = None
            elif mode == 'sparse':
                peak_engine = 'numpy'
                peak_method = 'geometrical'
                adjacency_radius_um = 450.
            
            cc.set_global_params(chunksize=1024,
                                            memory_mode=memory_mode,
                                            mode=mode,
                                            n_jobs=1,
                                            #~ adjacency_radius_um=adjacency_radius_um,
                                            )
            
            cc.set_preprocessor_params(
                    #signal preprocessor
                    highpass_freq=300,
                    lowpass_freq=5000.,
                    common_ref_removal=False,
                    smooth_size=0,
                    pad_width = None)
            
            
            cc.set_peak_detector_params(
                    #peak detector
                    method=peak_method, engine=peak_engine,
                    peak_sign='-', relative_threshold=7, peak_span_ms=0.5,
                    adjacency_radius_um=adjacency_radius_um,
                    )
            
            t1 = time.perf_counter()
            cc.estimate_signals_noise(seg_num=0, duration=10.)
            t2 = time.perf_counter()
            print('estimate_signals_noise', t2-t1)
            
            t1 = time.perf_counter()
            cc.run_signalprocessor(duration=10., detect_peak=True)
            t2 = time.perf_counter()
            print('run_signalprocessor_loop', t2-t1)

            for seg_num in range(dataio.nb_segment):
                mask = cc.all_peaks['segment']==seg_num
                print('seg_num', seg_num, 'nb peak',  np.sum(mask))
            
            # redetect peak
            cc.re_detect_peak(method=peak_method, engine=peak_engine,
                                                peak_sign='-', relative_threshold=5, peak_span_ms=0.7,
                                                adjacency_radius_um=adjacency_radius_um)
            for seg_num in range(dataio.nb_segment):
                mask = cc.all_peaks['segment']==seg_num
                print('seg_num', seg_num, 'nb peak',  np.sum(mask))

            cc.set_waveform_extractor_params(n_left=-25, n_right=40)

            t1 = time.perf_counter()
            cc.clean_peaks(alien_value_threshold=100, mode='extremum_amplitude')
            t2 = time.perf_counter()
            print('clean_peaks extremum_amplitude', t2-t1)

            t1 = time.perf_counter()
            cc.clean_peaks(alien_value_threshold=100, mode='full_waveform')
            t2 = time.perf_counter()
            print('clean_peaks full_waveforms', t2-t1)
            
            t1 = time.perf_counter()
            cc.sample_some_peaks( mode='rand', nb_max=5000)
            t2 = time.perf_counter()
            print('sample_some_peaks', t2-t1)
            
            
            
            print(cc)
            
            #extract_some_noise
            t1 = time.perf_counter()
            cc.extract_some_noise(nb_snippet=400)
            t2 = time.perf_counter()
            print('extract_some_noise', t2-t1)
            
            
            if mode == 'dense':
                # PCA
                t1 = time.perf_counter()
                cc.extract_some_features(method='global_pca', n_components=12)
                t2 = time.perf_counter()
                print('project pca', t2-t1)

                # cluster
                t1 = time.perf_counter()
                cc.find_clusters(method='kmeans', n_clusters=11)
                t2 = time.perf_counter()
                print('find_clusters', t2-t1)
            
            elif mode == 'sparse':
            
                # PCA
                t1 = time.perf_counter()
                cc.extract_some_features(method='pca_by_channel', n_components_by_channel=3)
                t2 = time.perf_counter()
                print('project pca', t2-t1)

                # cluster
                t1 = time.perf_counter()
                cc.find_clusters(method='pruningshears')
                t2 = time.perf_counter()
                print('find_clusters', t2-t1)
            
            print(cc)
        
        
        t1 = time.perf_counter()
        cc.cache_some_waveforms()
        t2 = time.perf_counter()
        print('cache_some_waveforms', t2-t1)
        
        t1 = time.perf_counter()
        cc.auto_split_cluster()
        t2 = time.perf_counter()
        print('auto_split_cluster', t2-t1)

        t1 = time.perf_counter()
        cc.trash_not_aligned()
        t2 = time.perf_counter()
        print('trash_not_aligned', t2-t1)

        t1 = time.perf_counter()
        cc.auto_merge_cluster()
        t2 = time.perf_counter()
        print('auto_merge_cluster', t2-t1)

        t1 = time.perf_counter()
        cc.trash_low_extremum()
        t2 = time.perf_counter()
        print('trash_low_extremum', t2-t1)


        t1 = time.perf_counter()
        cc.trash_small_cluster()
        t2 = time.perf_counter()
        print('trash_small_cluster', t2-t1)
        
        
            # similarity
            #~ cc.compute_centroid()
            #~ similarity, cluster_labels = cc.compute_similarity()
            #~ print(cluster_labels)
            #~ fig, ax = plt.subplots()
            #~ ax.matshow(similarity)
            #~ plt.show()
            
            #plot
            #~ wf = cc.some_waveforms
            #~ wf = wf.swapaxes(1,2).reshape(wf.shape[0], -1)
            
            #~ fig, ax = plt.subplots()
            #~ ax.plot(np.median(wf, axis=0), color='b')
            
            #~ plt.show()

    
    
    
def compare_nb_waveforms():
    if os.path.exists('test_catalogueconstructor'):
        shutil.rmtree('test_catalogueconstructor')
        
    dataio = DataIO(dirname='test_catalogueconstructor')
    localdir, filenames, params = download_dataset(name='olfactory_bulb')
    dataio.set_data_source(type='RawData', filenames=filenames, **params)
    dataio.add_one_channel_group(channels=range(14), chan_grp=0)


    cc = CatalogueConstructor(dataio=dataio)

    cc.set_global_params(chunksize=1024,
                                    memory_mode='ram',
                                    mode='dense',
                                    n_jobs=1,
                                    #~ adjacency_radius_um=None,
                                    )
    
    cc.set_preprocessor_params(
            #signal preprocessor
            highpass_freq=300,
            lowpass_freq=5000.,
            common_ref_removal=False,
            smooth_size=0,
            pad_width = None)
    
    
    cc.set_peak_detector_params(
            #peak detector
            method='global', engine='numpy',
            peak_sign='-', relative_threshold=7, peak_span_ms=0.5,
            #~ adjacency_radius_um=None,
            )
    
    t1 = time.perf_counter()
    cc.estimate_signals_noise(seg_num=0, duration=10.)
    t2 = time.perf_counter()
    print('estimate_signals_noise', t2-t1)
    
    t1 = time.perf_counter()
    cc.run_signalprocessor()
    t2 = time.perf_counter()
    print('run_signalprocessor', t2-t1)
    
    print(cc)
    
    fig, axs = plt.subplots(nrows=2)


    cc.set_waveform_extractor_params(wf_left_ms=-2.0, wf_right_ms=3.0)
    
    t1 = time.perf_counter()
    cc.sample_some_peaks( mode='rand', nb_max=5000)
    t2 = time.perf_counter()
    print('sample_some_peaks', t2-t1)
    
    colors = ['r', 'g', 'b', 'y']
    for i, nb_max in enumerate([100, 500, 1000, 2000]):
        
        cc.sample_some_peaks( mode='rand', nb_max=nb_max)
        #~ catalogueconstructor.extract_some_waveforms(wf_left_ms=-2.0, wf_right_ms=3.0,  nb_max=nb_max)
        #~ print(catalogueconstructor.some_waveforms.shape)
        t1 = time.perf_counter()
        wf = cc.get_some_waveforms()
        t2 = time.perf_counter()
        print('get_some_waveforms', nb_max,  t2-t1)
        
        #~ wf = catalogueconstructor.some_waveforms
        wf = wf.swapaxes(1,2).reshape(wf.shape[0], -1)
        axs[0].plot(np.median(wf, axis=0), color=colors[i], label='nb_max {}'.format(nb_max))
        
        axs[1].plot(np.mean(wf, axis=0), color=colors[i], label='nb_max {}'.format(nb_max))
    
    axs[0].legend()
    axs[0].set_title('median')
    axs[1].set_title('mean')
    plt.show()        




def test_make_catalogue():
    dataio = DataIO(dirname='test_catalogueconstructor')

    cc = CatalogueConstructor(dataio=dataio)

    #~ cc.make_catalogue()
    cc.make_catalogue_for_peeler()
    
    #~ for i, k in cc.catalogue['label_to_index'].items():
    
        #~ fig, ax = plt.subplots()
        #~ ax.plot(cc.catalogue['centers0'][i, :, :].T.flatten(), color='b')
        #~ ax.plot(cc.catalogue['centers1'][i, :, :].T.flatten(), color='g')
        #~ ax.plot(cc.catalogue['centers2'][i, :, :].T.flatten(), color='r')
        
    
    #~ plt.show()
    


def test_ratio_amplitude():
    dataio = DataIO(dirname='test_catalogueconstructor')
    catalogueconstructor = CatalogueConstructor(dataio=dataio)

    pairs = catalogueconstructor.detect_similar_waveform_ratio(0.5)
    print(pairs)


def test_create_savepoint_catalogue_constructor():
    dataio = DataIO(dirname='test_catalogueconstructor')
    catalogueconstructor = CatalogueConstructor(dataio=dataio)
    copy_path = catalogueconstructor.create_savepoint()
    print(copy_path)


def debug_interp_centers0():
    dataio = DataIO(dirname='test_catalogueconstructor')
    catalogueconstructor = CatalogueConstructor(dataio=dataio)
    catalogue = catalogueconstructor.make_catalogue()
    centers = catalogue['centers0']
    interp_centers = catalogue['interp_centers0']
    subsample_ratio = catalogue['subsample_ratio']
    
    #~ fig, ax = plt.subplots()
    #~ ax.plot(centers[0])
    #~ ax.plot(np.arange(interp_centers[0].shape[0])/subsample_ratio - 0.5, interp_centers[0])
    
    #~ plt.show()

def test_feature_with_lda_selection():
    dataio = DataIO(dirname='test_catalogueconstructor')
    cc = CatalogueConstructor(dataio=dataio)
    print(cc)
    
    selection = np.in1d(cc.all_peaks['cluster_label'], [1, 2, 3, 4])
    print(np.sum(selection), '/', cc.all_peaks.size)
    
    cc.extract_some_features(method='global_lda',  selection=selection)
    print(cc.some_features.shape)
    print(cc.some_features)
    
if __name__ == '__main__':
    test_catalogue_constructor()
    
    ## compare_nb_waveforms()
    
    #~ test_make_catalogue()
    #~ test_ratio_amplitude()
    
    #~ test_create_savepoint_catalogue_constructor()
    
    #~ debug_interp_centers0()
    
    #~ test_feature_with_lda_selection()





import os
from collections import OrderedDict

import numpy as np
import scipy.io

try:
    import pandas as pd
    HAS_PANDAS = True
except ImportError:
    HAS_PANDAS = False


class GenericSpikeExporter:
    def __call__(self,spikes, catalogue, seg_num, chan_grp, export_path,
                        split_by_cluster=False,
                        use_cell_label=True,
                        #~ use_index=True,
                        ):
        if not os.path.exists(export_path):
            os.makedirs(export_path)
        #~ print('export', spikes.size, seg_num, export_path)
        #~ print('split_by_cluster', split_by_cluster, 'use_cell_label', use_cell_label)
        clusters = catalogue['clusters']

        spike_labels = spikes['cluster_label']
        if use_cell_label:
            spike_labels = spikes['cluster_label'].copy()
            for l in clusters:
                mask = spike_labels==l['cluster_label']
                spike_labels[mask] = l['cell_label']

        spike_indexes = spikes['index']


        out_data = OrderedDict()
        if split_by_cluster:
            if use_cell_label:
                possible_labels = np.unique(clusters['cell_label'])
                label_name = 'cell'
            else:
                possible_labels = clusters['cluster_label']
                label_name = 'cluster'
            for k in possible_labels:
                keep = k == spike_labels
                out_data[label_name + '#'+ str(k)] = (spike_indexes[keep], spike_labels[keep])
        else:
            out_data['cell#all'] = (spike_indexes, spike_labels)

        name = 'spikes - segNum {} - chanGrp {}'.format(seg_num, chan_grp)
        filename = os.path.join(export_path, name)
        self.write_out_data(out_data, filename)

class CsvSpikeExporter(GenericSpikeExporter):
    ext = 'csv'
    def write_out_data(self, out_data, filename):
        for key, (spike_indexes, spike_labels) in out_data.items():
            filename2 = filename +' - '+key+'.csv'
            self._write_one_file(filename2, spike_indexes, spike_labels)

    def _write_one_file(self, filename, labels, indexes):
        rows = [''] * len(labels)
        for i in range(len(labels)):
            rows[i]='{},{}\n'.format(labels[i], indexes[i])
        with open(filename, 'w') as out:
            out.writelines(rows)

export_csv = CsvSpikeExporter()


class MatlabSpikeExporter(GenericSpikeExporter):
    ext = 'mat'

    def write_out_data(self, out_data, filename):
        mdict = {}

        for key, (spike_indexes, spike_labels) in out_data.items():
            mdict['index_'+key] = spike_indexes
            mdict['label_'+key] =spike_labels
        scipy.io.savemat(filename+'.mat', mdict)

export_matlab = MatlabSpikeExporter()


class ExcelSpikeExporter(GenericSpikeExporter):
    ext = 'xslx'
    def write_out_data(self, out_data, filename):
        assert HAS_PANDAS
        writer = pd.ExcelWriter(filename+'.xlsx')
        for key, (spike_indexes, spike_labels) in out_data.items():
            df = pd.DataFrame()
            df['index'] = spike_indexes
            df['label'] = spike_labels
            df.to_excel(writer, sheet_name=key, index=False)
        writer.save()

export_excel = ExcelSpikeExporter()


# list
export_list = [export_csv, export_matlab, ]
if HAS_PANDAS:
    export_list.append(export_excel)

export_dict = {e.ext:e for e in export_list}


def export_catalogue_spikes(cc, export_path=None, formats=None):
    """
    This export spikes from catalogue.
    Usefull when when catalogue peak sampler mode is all.
    This avoid the peeler.
    """
    
    dataio = cc.dataio
    chan_grp = cc.chan_grp
    
    sampler_mode = cc.info['peak_sampler']['mode']
    if sampler_mode != 'all':
        print('You are trying to export peak from catalogue but peak_sampler mode is not "all"')

    if export_path is None:
        export_path = os.path.join(dataio.dirname, 'export_catalogue_chan_grp_{}'.format(chan_grp))
    
    catalogue = {}
    catalogue['clusters'] = cc.clusters.copy()
    
    if formats is None:
        exporters = export_list
    elif isinstance(formats, str):
        assert formats in export_dict
        exporters = [ export_dict[formats] ]
    elif isinstance(format, list):
        exporters = [ export_dict[format] for format in formats]
    else:
        raise ValueError()
        
    for seg_num in range(dataio.nb_segment):
        in_segment = (cc.all_peaks['segment'] == seg_num)
        pos_label = (cc.all_peaks['cluster_label'] >= 0)
        spikes = cc.all_peaks[in_segment & pos_label]
        
        if spikes is None: continue
        
        args = (spikes, catalogue, seg_num, chan_grp, export_path,)
        kargs = dict(split_by_cluster=False, use_cell_label=False)
        for exporter in exporters:
            exporter(*args, **kargs)



"""
This should be rewritte with vispy but I don't have time now...
"""
from .myqt import QT
import pyqtgraph as pg

from matplotlib.path import Path as mpl_path

import numpy as np
import pandas as pd

import itertools

from .base import WidgetBase
from .tools import ParamDialog
from ..tools import median_mad, get_neighborhood
from .. import labelcodes


class MyViewBox(pg.ViewBox):
    doubleclicked = QT.pyqtSignal()
    gain_zoom = QT.pyqtSignal(float)
    #~ xsize_zoom = QT.pyqtSignal(float)
    #~ lasso_started = QT.pyqtSignal()
    lasso_drawing = QT.pyqtSignal(object)
    lasso_finished = QT.pyqtSignal(object)
    
    def __init__(self, *args, **kwds):
        pg.ViewBox.__init__(self, *args, **kwds)
        self.disableAutoRange()
        self.drag_points = []
        
    def mouseClickEvent(self, ev):
        ev.accept()
        
    def mouseDoubleClickEvent(self, ev):
        self.doubleclicked.emit()
        ev.accept()
        
    def wheelEvent(self, ev, axis=None):
        if ev.modifiers() == QT.Qt.ControlModifier:
            z = 10 if ev.delta()>0 else 1/10.
        else:
            z = 1.3 if ev.delta()>0 else 1/1.3
        self.gain_zoom.emit(z)
        ev.accept()
        
    def mouseDragEvent(self, ev):
        ev.accept()
        if ev.button()!=1: return
        
        if ev.isStart():
            self.drag_points = []
        
        pos = self.mapToView(ev.pos())
        self.drag_points.append([pos.x(), pos.y()])
        
        if ev.isFinish():
            self.lasso_finished.emit(self.drag_points)
        else:
            self.lasso_drawing.emit(self.drag_points)
        

class NDScatter(WidgetBase):
    """
    ND Scatter a central beast for the catalogue window.
    
    It try to mimic `RGGobi viewer package <http://www.ggobi.org/rggobi/>`_.
    
    Displaying clusters in high (more than 3!) dimensional space is difficult.
    To overcome this problem, some people use a combination of 2D plots 
    of all pairs of dimensions. Others use various 3D representations.
    However in this two cases the view angle is seriously restricted: 
    it is the projection of a big hyper cube (ND) onto only one face (2D or 3D)
    of this hyper cube. In real life, two clusters are not necessary well separated
    in one of this hyper-cube's face but better in a complex hyper plane that brings
    into play a combination of all features/dimensions.
    
    ND Scatter project randomly N-D space into 2-D space. Furthermore it do
    it dynamically. This is the **random tour**. 
    
    If you want to compare how 2 by 2 components are displayed in
    other tools you can use **next face**. This will combined (PC0, PC1), then 
    (PC0, PC2), ...
    
    For very high dimention, ND Scatter is not enough, for so ND scatter is guided in feature
    selection with the probe geometry. In short when you set visible some label with the
    **Cluster list** this automatically activated the dimenssions that must but 
    visible (and rodomized) and hidden. This is based on **extremum_channel**
    property of clusters. This is the **auto_select_component** in settings. Note 
    that you can define a radius in  micometers arround the channel of the max so
    that all component in a neighborhood of thsi channel will be also displayed.
    this is based on the PRB file that **need to have the units in  μm** (contrary to
    phy where units is not important for plotting)
    
    Note:
      * for too dense cluster you can limit the nb of dot dsiplayed.
        It helps a lot when density of clusters are differents.
        The is the role of **random decimate** button
      * you can manually activate/deactivate componant just for to understand.
      * mouse wheel will zoom both X et Y
      * with right click you can draw a **lasso**. This will select peak inside
        the poly on the projected hyperplan. This is in fact the secret weappon
        of tridesclous. It help a lot for au zoom on trace for "strange" point ouside
        main clouds. It can help to understand if they are in fact superposition of
        several spikes or very few dense cluster.
      * not all peak can be visible here only thosse wich have a waveform and so a
        feature. remember that not all waveforms are taken for clustering but only
        a subset. Getting everything is the role of the Peeler.
      * This viewer take some CPU ressource since in **random tour** a numpy.dot
        is done for each step.

    """
    _params = [{'name': 'refresh_interval', 'type': 'float', 'value': 100 },
                       {'name': 'nb_step', 'type': 'int', 'value':  10, 'limits' : [5, 100] },
                       {'name': 'max_visible_by_cluster', 'type': 'int', 'value':  1000, 'limits' : [10, 10000], 'step':50 },
                       {'name': 'auto_select_component', 'type': 'bool', 'value': True},
                       {'name': 'radius_um', 'type': 'float', 'value': 300.},
                   ]
    
    def __init__(self, controller=None, parent=None):
        WidgetBase.__init__(self, parent=parent, controller=controller)
        
        self.layout = QT.QHBoxLayout()
        self.setLayout(self.layout)
        
        self.create_toolbar()
        
        self.graphicsview = pg.GraphicsView()
        self.layout.addWidget(self.graphicsview)

        self.toolbar.addStretch()
        self.graphicsview2 = pg.GraphicsView()
        self.toolbar.addWidget(self.graphicsview2)

        #~ _params = [{'name': 'refresh_interval', 'type': 'float', 'value': 100 },
                           #~ {'name': 'nb_step', 'type': 'int', 'value':  10, 'limits' : [5, 100] },
                           #~ {'name': 'max_visible_by_cluster', 'type': 'int', 'value':  1000, 'limits' : [10, 10000], 'step':50 },
                           #~ ]
        #~ self.params = pg.parametertree.Parameter.create( name='Global options', type='group', children = _params)
        #~ self.tree_params = pg.parametertree.ParameterTree(parent  = self)
        #~ self.tree_params.header().hide()
        #~ self.tree_params.setParameters(self.params, showTop=True)
        #~ self.tree_params.setWindowTitle(u'Options for NDScatter')
        #~ self.tree_params.setWindowFlags(QT.Qt.Window)

        
        self.timer_tour = QT.QTimer(interval = 100)
        self.timer_tour.timeout.connect(self.new_tour_step)
        
        #~ if self.data is not None:
        self.initialize()
        self.refresh()
    
    def create_toolbar(self):
        
        tb = self.toolbar = QT.QVBoxLayout()
        self.layout.addLayout(tb)
        but = QT.QPushButton('next face')
        tb.addWidget(but)
        but.clicked.connect(self.next_face)
        but = QT.QPushButton('Random')
        tb.addWidget(but)
        but.clicked.connect(self.random_projection)
        but = QT.QPushButton('Random tour', checkable = True)
        tb.addWidget(but)
        but.clicked.connect(self.start_stop_tour)
        but = QT.QPushButton('settings')
        but.clicked.connect(self.open_settings)
        tb.addWidget(but)
        but = QT.QPushButton('random decimate', icon=QT.QIcon.fromTheme("roll"))
        but.clicked.connect(self.by_cluster_random_decimate)
        tb.addWidget(but)
        but = QT.QPushButton('select component')
        but.clicked.connect(self.open_select_component)
        tb.addWidget(but)
        
    def open_select_component(self):
        
        ndim = self.data.shape[1]
        params = [{'name' : 'comp {}'.format(i), 'type':'bool', 'value' : self.selected_comp[i]} for i in range(ndim)]
        
        dialog = ParamDialog(params, title = 'Select component')
        if dialog.exec_():
            p = dialog.get()
            for i in range(ndim):
                self.selected_comp[i] = p['comp {}'.format(i)]
            self.refresh()
    
    # this handle data with propties so model change shoudl not affect so much teh code
    @property
    def data(self):
        
        feat = self.controller.some_features
        data = ensure_2d(feat)
        return data
    
    def data_by_label(self, k):
        if len(self.point_visible) != self.data.shape[0]:
            self.point_visible = np.zeros(self.data.shape[0], dtype=bool)
            self.by_cluster_random_decimate()
        
        if k=='sel':
            data = self.data[self.controller.spike_selection[self.controller.some_peaks_index]]
        elif k==labelcodes.LABEL_NOISE:
            feat_noise = self.controller.some_noise_features
            data = ensure_2d(feat_noise)
        else:
            data = self.data[(self.controller.spike_label[self.controller.some_peaks_index]==k) & self.point_visible]
        
        return data
    
    def by_cluster_random_decimate(self, clicked=None, refresh=True):
        m = self.params['max_visible_by_cluster']
        for k in self.controller.cluster_labels:
            mask = self.controller.spike_label[self.controller.some_peaks_index]==k
            if self.controller.cluster_count[k]>m:
                self.point_visible[mask] = False
                visible, = np.nonzero(mask)
                if visible.size>0:
                    visible = np.random.choice(visible, size=m)
                    self.point_visible[visible] = True
            else:
                self.point_visible[mask] = True
        
        if refresh:
            self.refresh()
        
    def get_color(self, k):
        color = self.controller.qcolors.get(k, QT.QColor( 'white'))
        return color
    
    def is_cluster_visible(self, k):
        return self.controller.cluster_visible[k]

    
    def initialize(self):
        #~ if self.data is None:
        if self.controller.some_features is None:
            return
        self.viewBox = MyViewBox()
        self.viewBox.gain_zoom.connect(self.gain_zoom)
        #~ self.viewBox.lasso_started.connect(self.on_lasso_started)
        self.viewBox.lasso_drawing.connect(self.on_lasso_drawing)
        self.viewBox.lasso_finished.connect(self.on_lasso_finished)
        self.plot = pg.PlotItem(viewBox=self.viewBox)
        self.graphicsview.setCentralItem(self.plot)
        self.plot.hideButtons()
        
        self.scatter = pg.ScatterPlotItem(size=3, pxMode = True)
        self.plot.addItem(self.scatter)
        self.scatter.sigClicked.connect(self.on_scatter_clicked)
        
        
        brush = QT.QColor( 'magenta')
        brush.setAlpha(180)
        self.scatter_select = pg.ScatterPlotItem(pen=pg.mkPen(None), brush=brush, size=11, pxMode = True)
        self.plot.addItem(self.scatter_select)
        self.scatter_select.setZValue(1000)
        
        
        color = self.controller.qcolors.get(labelcodes.LABEL_NOISE)
        self.scatter_noise = pg.ScatterPlotItem(pen=pg.mkPen(None), brush=color, size=3, pxMode = True)
        self.plot.addItem(self.scatter_noise)
        
        self.lasso = pg.PlotCurveItem(pen='#7FFF00')
        self.plot.addItem(self.lasso)
        
        #estimate limts
        #  VERY SLOW
        #~ med, mad = median_mad(self.data)
        #~ m = 4.*np.max(mad)
        #~ self.limit = m
        #~ self.plot.setXRange(-m, m)
        #~ self.plot.setYRange(-m, m)
        
        #estimate limts
        data = self.data.flatten()
        if data.size > 1000:
            data = data.take(np.random.choice(data.size, 1000, replace=False))
        min_ = np.min(data)
        max_ = np.max(data)
        m = max(np.abs(min_), np.abs(max_)) * 2.5
        self.limit = m
        
        ndim = self.data.shape[1]
        if ndim <2:
            ndim = 2
        self.selected_comp = np.ones( (ndim), dtype='bool')
        self.projection = np.zeros( (ndim, 2))
        self.projection[0,0] = 1.
        self.projection[1,1] = 1.
        
        self.point_visible = np.zeros(self.data.shape[0], dtype=bool)
        self.by_cluster_random_decimate(refresh=False)
        
        self.plot2 = pg.PlotItem(viewBox=MyViewBox(lockAspect=True))
        self.graphicsview2.setCentralItem(self.plot2)
        self.plot2.hideButtons()
        angles = np.arange(0,360, .1)
        self.circle = pg.PlotCurveItem(x=np.cos(angles), y=np.sin(angles), pen=(255,255,255))
        self.plot2.addItem(self.circle)
        self.direction_lines = pg.PlotCurveItem(x=[], y=[], pen=(255,255,255))
        self.direction_data = np.zeros( (ndim*2, 2))
        self.plot2.addItem(self.direction_lines)
        self.plot2.setXRange(-1, 1)
        self.plot2.setYRange(-1, 1)
        self.proj_labels = []
        for i in range(ndim):
            text = 'PC{}'.format(i)
            label = pg.TextItem(text, color=(1,1,1), anchor=(0.5, 0.5), border=None, fill=pg.mkColor((128,128,128, 180)))
            self.proj_labels.append(label)
            self.plot2.addItem(label)
        
        self.graphicsview2.setMaximumSize(200, 200)
        
        #~ self.hyper_faces = list(itertools.product(range(ndim), range(ndim)))
        self.hyper_faces = list(itertools.permutations(range(ndim), 2))
        self.n_face = -1
    
    def next_face(self):
        self.n_face += 1
        self.n_face = self.n_face%len(self.hyper_faces)
        ndim = self.data.shape[1]
        self.projection = np.zeros( (ndim, 2))
        i, j = self.hyper_faces[self.n_face]
        self.projection[i,0] = 1.
        self.projection[j,1] = 1.
        if self.timer_tour.isActive():
            self.tour_step = 0
        self.refresh()
        
    def get_one_random_projection(self):
        ndim = self.data.shape[1]
        projection = np.random.rand(ndim,2)*2-1.
        projection[~self.selected_comp] = 0
        m = np.sqrt(np.sum(projection**2, axis=0))
        ok = m > 0
        projection[:, ok] /= m[ok]
        return projection
    
    def random_projection(self):
        self.projection = self.get_one_random_projection()
        if self.timer_tour.isActive():
            self.tour_step == 0
        self.refresh()
    
    def apply_dot(self, data):
        #~ print(data.shape, self.projection.shape)
        projected = np.dot(data[:, self.selected_comp ], self.projection[self.selected_comp, :])
        return projected
    
    def refresh(self):
        #~ if self.data is None:
        if self.controller.some_features is None:
            if hasattr(self, 'plot'):
                self.plot.clear()
            return

        if not hasattr(self, 'viewBox'):
            self.initialize()
        
        if self.data.shape[1] != self.projection.shape[0]:
            self.initialize()
        
        #ndscatter
        self.scatter.clear()

        for k in self.controller.cluster_labels:
            if not self.is_cluster_visible(k): continue
            data = self.data_by_label(k)
            #~ projected = np.dot(data, self.projection )
            projected = self.apply_dot(data)
            color = self.get_color(k)
            self.scatter.addPoints(x=projected[:,0], y=projected[:,1],  pen=pg.mkPen(None), brush=color)
        
        #selection scatter
        data_sel = self.data_by_label('sel')
        #~ projected = np.dot(data_sel, self.projection )
        projected = self.apply_dot(data_sel)
        self.scatter_select.setData(projected[:,0], projected[:,1])
        
        #noise
        if self.is_cluster_visible(labelcodes.LABEL_NOISE):
            data_noise = self.data_by_label(labelcodes.LABEL_NOISE)
            if data_noise is not None:
                projected = self.apply_dot(data_noise)
                self.scatter_noise.setData(projected[:,0], projected[:,1])
                self.scatter_noise.show()
        else:
            self.scatter_noise.hide()
        
        #projection axes
        proj = self.projection.copy()
        proj[~self.selected_comp, :] = 0
        self.direction_data[::, :] =0
        #~ self.direction_data[::2, :] = self.projection
        self.direction_data[::2, :] = proj
        self.direction_lines.setData(self.direction_data[:,0], self.direction_data[:,1])
        
        for i, label in enumerate(self.proj_labels):
            if self.selected_comp[i]:
                label.setPos(self.projection[i,0], self.projection[i,1])
                label.show()
            else:
                label.hide()
        
        self.graphicsview.repaint()
            
    
    def start_stop_tour(self, checked):
        if checked:
            self.tour_step = 0
            self.timer_tour.setInterval(self.params['refresh_interval'])
            self.timer_tour.start()
        else:
            self.timer_tour.stop()
    
    def new_tour_step(self):
        #~ if self.data is None:
        if self.controller.some_features is None:
            return
        nb_step = self.params['nb_step']
        ndim = self.data.shape[1]
        
        if self.tour_step == 0:
            self.tour_steps = np.empty( (ndim , 2 ,  nb_step))
            arrival = self.get_one_random_projection()
            for i in range(ndim):
                for j in range(2):
                    self.tour_steps[i,j , : ] = np.linspace(self.projection[i,j] , arrival[i,j] , nb_step)
            m = np.sqrt(np.sum(self.tour_steps**2, axis=0))
            m = m[np.newaxis, : ,  :]
            self.tour_steps /= m
        
        self.projection = self.tour_steps[:,:,self.tour_step]
        
        self.tour_step+=1
        if self.tour_step>=nb_step:
            self.tour_step = 0
            
        self.refresh()

    def gain_zoom(self, factor):
        self.limit /= factor
        l = float(self.limit)
        self.plot.setXRange(-l, l)
        self.plot.setYRange(-l, l)
    
    def on_scatter_clicked(self,plots, points):
        self.controller.spike_selection[:] = False
        if len(points)==1:
            #~ projected = np.dot(self.data, self.projection )
            projected = self.apply_dot(self.data)
            pos = points[0].pos()
            pos = [pos.x(), pos.y()]
            ind = np.argmin(np.sum((projected-pos)**2, axis=1))
            self.controller.spike_selection[self.controller.some_peaks_index[ind]] = True
            
        self.refresh()
        self.spike_selection_changed.emit()
    
    def on_lasso_drawing(self, points):
        points = np.array(points)
        self.lasso.setData(points[:, 0], points[:, 1])
    
    def on_lasso_finished(self, points):
        self.lasso.setData([], [])
        vertices = np.array(points)
        
        self.controller.spike_selection[:] = False
        #~ projected = np.dot(self.data, self.projection )
        projected = self.apply_dot(self.data)
        inside = inside_poly(projected, vertices)
        visibles = self.controller.spike_visible[self.controller.some_peaks_index]
        self.controller.spike_selection[self.controller.some_peaks_index[inside&visibles]] = True
        self.refresh()
        
        self.spike_selection_changed.emit()
    
    def auto_select_component(self):
        #~ if self.data is None:
        if self.controller.some_features is None:
            return
        
        # take all component that corespon to a channel max and neighborhood
        neighborhood = get_neighborhood(self.controller.geometry, self.params['radius_um'])
        
        ndim = self.data.shape[1]
        self.selected_comp = np.zeros( (ndim), dtype='bool')
        for k, v in self.controller.cluster_visible.items():
            if not v: continue
            c = self.controller.get_extremum_channel(k)
            if c is not None:
                chan_mask = neighborhood[c,:]
                #~ print('c', c, chan_mask)
                feat_mask = np.sum(self.controller.channel_to_features[chan_mask, :], axis=0).astype('bool')
                #~ print(feat_mask)
                
                self.selected_comp |= feat_mask
        
        #~ print('auto_select_component', self.selected_comp)

    #~ def on_spike_selection_changed(self):
        #~ self.refresh()

    #~ def on_spike_label_changed(self):
        #~ self.refresh()
        
    #~ def on_colors_changed(self):
        #~ self.refresh()
    
    def on_cluster_visibility_changed(self):
        #~ if self.data is None:
        if self.controller.some_features is None:
            return
        
        if self.params['auto_select_component']:
            self.auto_select_component()
        #~ self.refresh()
        self.random_projection()




def inside_poly(data, vertices):
    return mpl_path(vertices).contains_points(data)


def ensure_2d(feat):
    if feat is None:
        return None
    # ensure 2D
    if feat.shape[1] ==1:
        data = np.zeros((feat.shape[0], 2), dtype=feat.dtype)
        data[:, 0] = feat[:, 0]
    else:
        data = feat
    return data
    



import numpy as np

import matplotlib.cm
import matplotlib.colors

from ..gui import QT
import pyqtgraph as pg
from pyqtgraph.util.mutex import Mutex

import pyacq
from pyacq import WidgetNode,ThreadPollInput, StreamConverter, InputStream


#~ _dtype_spike = [('index', 'int64'), ('label', 'int64'), ('jitter', 'float64'),]
from ..peeler import _dtype_spike
from ..tools import make_color_dict
from ..labelcodes import LABEL_UNCLASSIFIED

import time

class MyViewBox(pg.ViewBox):
    doubleclicked = QT.pyqtSignal()
    gain_zoom = QT.pyqtSignal(float)
    def mouseDoubleClickEvent(self, ev):
        self.doubleclicked.emit()
        ev.accept()
    def wheelEvent(self, ev, axis=None):
        if ev.modifiers() == QT.Qt.ControlModifier:
            z = 10 if ev.delta()>0 else 1/10.
        else:
            z = 1.3 if ev.delta()>0 else 1/1.3
        self.gain_zoom.emit(z)
        ev.accept()
    def raiseContextMenu(self, ev):
        #for some reasons enableMenu=False is not taken (bug ????)
        pass



    
class OnlineWaveformHistViewer(WidgetNode):
    
    _input_specs = {'signals': dict(streamtype='signals'),
                                'spikes': dict(streamtype='events', shape = (-1, ),  dtype=_dtype_spike),
                                    }
    
    _params = [
                      {'name': 'colormap', 'type': 'list', 'values' : ['hot', 'viridis', 'jet', 'gray',  ] },
                      {'name': 'bin_min', 'type': 'float', 'value' : -20. },
                      {'name': 'bin_max', 'type': 'float', 'value' : 8. },
                      {'name': 'bin_size', 'type': 'float', 'value' : .1 },
                      {'name': 'refresh_interval', 'type': 'int', 'value': 100, 'limits':[5, 1000]},
                      ]

    
    def __init__(self, **kargs):
        WidgetNode.__init__(self, **kargs)

        self.layout = QT.QVBoxLayout()
        self.setLayout(self.layout)
        
        
        h = QT.QHBoxLayout()
        self.layout.addLayout(h)
        
        self.combobox = QT.QComboBox()
        h.addWidget(self.combobox)
        
        but = QT.QPushButton('clear')
        h.addWidget(but)
        but.clicked.connect(self.on_clear)
        
        self.label = QT.QLabel('')
        h.addWidget(self.label)
        
        self.graphicsview = pg.GraphicsView()
        self.layout.addWidget(self.graphicsview)

        self.params = pg.parametertree.Parameter.create( name='settings', type='group', children=self._params)
        self.tree_params = pg.parametertree.ParameterTree(parent=self)
        self.tree_params.header().hide()
        self.tree_params.setParameters(self.params, showTop=True)
        self.tree_params.setWindowTitle('Options for waveforms hist viewer')
        self.tree_params.setWindowFlags(QT.Qt.Window)
        self.params.sigTreeStateChanged.connect(self.on_params_changed)
        
        self.initialize_plot()
        
        self.mutex = Mutex()

        
        
    def _configure(self, peak_buffer_size=100000, catalogue=None, **kargs):
        self.peak_buffer_size = peak_buffer_size
        self.catalogue = catalogue
    
    
    def _initialize(self, **kargs):

        self.sample_rate =  self.inputs['signals'].params['sample_rate']
        self.wf_dtype =  self.inputs['signals'].params['dtype']
        
        self.inputs['spikes'].set_buffer(size=self.peak_buffer_size, double=False)
        buffer_sigs_size = int(self.sample_rate*3.)
        self.inputs['signals'].set_buffer(size=buffer_sigs_size, double=False)

        # poller
        self.poller_sigs = ThreadPollInput(input_stream=self.inputs['signals'], return_data=True)
        self.poller_spikes = ThreadPollInput(input_stream=self.inputs['spikes'], return_data=True)

        self.histogram_2d = {}
        self.last_waveform = {}
        self.change_catalogue(self.catalogue)
        
        self.timer = QT.QTimer(interval=100)
        self.timer.timeout.connect(self.refresh)
    
    def _start(self, **kargs):
        self.last_head_sigs = None
        self.last_head_spikes = None
        self.timer.start()
        self.inputs['signals'].empty_queue()
        self.inputs['spikes'].empty_queue()
        self.poller_sigs.start()
        self.poller_spikes.start()

    def _stop(self, **kargs):
        self.timer.stop()
        self.poller_sigs.stop()
        self.poller_sigs.wait()
        self.poller_spikes.stop()
        self.poller_spikes.wait()

    def _close(self, **kargs):
        pass

    def open_settings(self):
        if not self.tree_params.isVisible():
            self.tree_params.show()
        else:
            self.tree_params.hide()
    
    def on_params_changed(self, params, changes):
        self.change_lut()
        self.change_catalogue(self.catalogue)
        self.timer.setInterval(self.params['refresh_interval'])

    def initialize_plot(self):
        
        self.viewBox = MyViewBox()
        self.viewBox.doubleclicked.connect(self.open_settings)
        self.viewBox.gain_zoom.connect(self.gain_zoom)
        self.viewBox.disableAutoRange()
        
        self.plot = pg.PlotItem(viewBox=self.viewBox)
        self.graphicsview.setCentralItem(self.plot)
        self.plot.hideButtons()
        
        self.image = pg.ImageItem()
        self.plot.addItem(self.image)
        
        self.curve_spike = pg.PlotCurveItem()
        self.plot.addItem(self.curve_spike)

        self.curve_limit = pg.PlotCurveItem()
        self.plot.addItem(self.curve_limit)
        
        self.change_lut()


    def change_lut(self):
        N = 512
        cmap_name = self.params['colormap']
        cmap = matplotlib.cm.get_cmap(cmap_name , N)
        lut = []
        for i in range(N):
            r,g,b,_ =  matplotlib.colors.ColorConverter().to_rgba(cmap(i))
            lut.append([r*255,g*255,b*255])
        self.lut = np.array(lut, dtype='uint8')

    def change_catalogue(self, catalogue):
        self.params.blockSignals(True)
        with self.mutex:
            
            self.catalogue = catalogue
            
            colors = make_color_dict(self.catalogue['clusters'])
            self.qcolors = {}
            for k, color in colors.items():
                r, g, b = color
                self.qcolors[k] = QT.QColor(r*255, g*255, b*255)

            self.all_plotted_labels = self.catalogue['cluster_labels'].tolist() + [LABEL_UNCLASSIFIED]
            
            centers0 = self.catalogue['centers0']
            if centers0.shape[0]>0:
                self.params['bin_min'] = np.min(centers0)*1.5
                self.params['bin_max'] = np.max(centers0)*1.5
            
            
            bin_min, bin_max = self.params['bin_min'], self.params['bin_max']
            bin_size = self.params['bin_size']
            self.bins = np.arange(bin_min, bin_max, self.params['bin_size'])

            self.combobox.clear()
            self.combobox.addItems([str(k) for k in self.all_plotted_labels])
        
        self.on_clear()
        self._max = 10
        
        _, peak_width, nb_chan = self.catalogue['centers0'].shape
        x, y = [], []
        for c in range(1, nb_chan):
            x.extend([c*peak_width, c*peak_width, np.nan])
            y.extend([-1000, 1000, np.nan])
        
        self.curve_limit.setData(x=x, y=y, connect='finite')
        self.params.blockSignals(False)
        
        

    def on_clear(self):
        with self.mutex:
            shape = self.catalogue['centers0'].shape
            
            self.indexes0 = np.arange(shape[1]*shape[2], dtype='int64')
            
            self.histogram_2d = {}
            self.last_waveform = {}
            self.nb_spikes = {}
            for k in self.all_plotted_labels:
                self.histogram_2d[k] = np.zeros((shape[1]*shape[2], self.bins.size), dtype='int64')
                self.last_waveform[k] = np.zeros((shape[1]*shape[2],), dtype=self.wf_dtype)
                self.nb_spikes[k] = 0
        
        self.plot.setXRange(0, self.indexes0[-1]+1)
        self.plot.setYRange(self.params['bin_min'], self.params['bin_max'])
    
    def auto_scale(self):
        pass
        

    def gain_zoom(self, v):
        self._max *= v
        self.image.setLevels([0, self._max], update=True)
    
    def refresh(self):
        #~ print('refresh')
        #~ t0 = time.perf_counter()
        
        head_sigs = self.poller_sigs.pos()
        head_spikes = self.poller_spikes.pos()
        
        if self.last_head_sigs is None:
            self.last_head_sigs = head_sigs
        
        if self.last_head_spikes is None:
            self.last_head_spikes = head_spikes
        
        if self.last_head_spikes is None or self.last_head_sigs is None:
            return
        
        # update image
        n_right, n_left = self.catalogue['n_right'],self.catalogue['n_left']
        bin_min, bin_max, bin_size = self.params['bin_min'], self.params['bin_max'],self.params['bin_size']
        
        
        # check peak_buffer_size here
        if (head_spikes-self.last_head_spikes)>=(0.9*self.peak_buffer_size):
            self.last_head_spikes = head_spikes - int(0.9*self.peak_buffer_size)
        new_spikes = self.inputs['spikes'].get_data(self.last_head_spikes, head_spikes)
        
        right_indexes = new_spikes['index'] + n_right
        if np.any(right_indexes > head_sigs):
            # the buffer of signals is available for some spikes yet
            # so remove then for this loop and get back on head_spikes
            first_out = np.nonzero(right_indexes)[0][0]
            head_spikes = head_spikes - (new_spikes.size - first_out)
            new_spikes = new_spikes[:first_out]
        
        for k in self.all_plotted_labels:
            mask = new_spikes['cluster_label'] == k
            indexes = new_spikes[mask]['index']
            for ind in indexes:
                wf = self.inputs['signals'].get_data(ind+n_left, ind+n_right)
                wf = wf.T.reshape(-1)
                wf_bined = np.floor((wf-bin_min)/bin_size).astype('int32')
                wf_bined = wf_bined.clip(0, self.bins.size-1)
                
                with self.mutex:
                    self.histogram_2d[k][self.indexes0, wf_bined] += 1
                    self.last_waveform[k] = wf
                    self.nb_spikes[k] += 1
        
        self.last_head_sigs = head_sigs
        self.last_head_spikes = head_spikes

        
        if self.combobox.currentIndex() == -1:
            return

        if self.visibleRegion().isEmpty():
            # when several tabs not need to refresh
            return
        
        # refresh plot , update image
        k = self.all_plotted_labels[self.combobox.currentIndex()]
        hist2d = self.histogram_2d[k]

        self.image.setImage(hist2d, lut=self.lut, levels=[0, self._max])
        self.image.setRect(QT.QRectF(-0.5, bin_min, hist2d.shape[0], bin_max-bin_min))
        self.image.show()
        
        self.curve_spike.setData(x=self.indexes0, y=self.last_waveform[k],
                                            pen=pg.mkPen(self.qcolors[k], width=1.5))
        
        
        txt = 'nbs_pike = {}'.format(self.nb_spikes[k])
        self.label.setText(txt)
        
        #~ t1 = time.perf_counter()
        #~ print('refresh time', t1-t0)
        
        
    



import os, shutil

import numpy as np

from tridesclous.export import export_csv, export_matlab, export_excel, export_catalogue_spikes
from tridesclous.catalogueconstructor import _dtype_cluster
from tridesclous.peeler import _dtype_spike

from tridesclous.dataio import DataIO
from tridesclous.catalogueconstructor import CatalogueConstructor



from tridesclous.tests.testingtools import setup_catalogue

def setup_module():
    setup_catalogue('test_export_catalogue_spike', dataset_name='olfactory_bulb', peak_sampler_mode='all')


def test_export():
    if os.path.exists('test_export'):
        shutil.rmtree('test_export')
    
    export_path = 'test_export'
    nspike = 50
    ncluster = 3
    
    seg_num = 0
    chan_grp = 0
    
    clusters = np.zeros(ncluster, dtype=_dtype_cluster)
    clusters['cluster_label'] = [0, 1, 4]
    clusters['cell_label'] = [0, 1, 1]
    
    catalogue = {}
    catalogue['clusters'] = clusters
    
    spikes = np.zeros(nspike, dtype=_dtype_spike)
    spikes['index'] = np.sort(np.random.randint(0, high=1000000, size=nspike))
    spikes['cluster_label'] = clusters['cluster_label'][np.random.randint(0, high=ncluster, size=nspike)]
    
    #~ print(clusters)
    #~ print(spikes)
    
    for split_by_cluster in (True, False):
        for use_cell_label in (True, False):
            
            args = (spikes, catalogue, seg_num, chan_grp, export_path+'/split{} celllabel{}/'.format(split_by_cluster, use_cell_label),)
            kargs = dict(split_by_cluster=split_by_cluster, use_cell_label=use_cell_label)
            
            export_csv(*args, **kargs)
            export_matlab(*args, **kargs)
            export_excel(*args, **kargs)


def test_export_catalogue_spikes():
    dataio = DataIO(dirname='test_export_catalogue_spike')
    cc = CatalogueConstructor(dataio=dataio)
    #~ print(cc)
    
    export_catalogue_spikes(cc)
    
    
    
if __name__ == '__main__':
    setup_module()
    test_export()
    test_export_catalogue_spikes()




import numpy as np
import time


from tridesclous.waveformtools import extract_chunks

#~ size = 1000000
size = 1000
nb_channel = 5
#~ nb_channel = 300
width = 150
nb_peak = 20003


def test_extract_chunks_memory():
    signals = np.random.randn(size, nb_channel).astype('float32')
    indexes = np.random.randint(low=width, high=size-width, size=nb_peak)

    t0 = time.perf_counter()
    chunks = extract_chunks(signals, indexes, width, chunks=None)
    t1 = time.perf_counter()
    print('extract_chunks no buffer', t1-t0)
    

    chunks[:] = 0
    
    t0 = time.perf_counter()
    chunks = extract_chunks(signals, indexes, width, chunks=chunks)
    t1 = time.perf_counter()
    print('extract_chunks with buffer', t1-t0)




def test_extract_chunks_memmap():
    signals = np.memmap('test_extract_wf_signal', dtype='float32', mode='w+', shape=(size, nb_channel))
    indexes = np.random.randint(low=width, high=size-width, size=nb_peak)
    chunks = np.memmap('test_extract_wf_chunks', dtype='float32', mode='w+', shape=(nb_peak, width, nb_channel))
    
    t0 = time.perf_counter()
    chunks = extract_chunks(signals, indexes, width, chunks=chunks)
    chunks.flush()
    t1 = time.perf_counter()
    print('extract_chunks memmap to memmap', t1-t0)


def test_extract_chunks_with_channel_indexes():
    signals = np.random.randn(size, nb_channel).astype('float32')
    indexes = np.random.randint(low=width, high=size-width, size=nb_peak)

    t0 = time.perf_counter()
    chunks = extract_chunks(signals, indexes, width, channel_indexes=[0,2,3], chunks=None)
    t1 = time.perf_counter()
    print('extract_chunks channel_indexes', t1-t0)



    

if __name__ == '__main__':
    test_extract_chunks_memory()
    test_extract_chunks_memmap()
    test_extract_chunks_with_channel_indexes()
    
    
    





from .myqt import QT
import pyqtgraph as pg

import numpy as np
import matplotlib.cm
import matplotlib.colors


from .base import WidgetBase
from .tools import ParamDialog
from ..tools import median_mad

import time

#~ enableMenu
class MyViewBox(pg.ViewBox):
    doubleclicked = QT.pyqtSignal()
    gain_zoom = QT.pyqtSignal(float)
    def mouseDoubleClickEvent(self, ev):
        self.doubleclicked.emit()
        ev.accept()
    #~ def wheelEvent(self, ev, axis=None):
        #~ if ev.modifiers() == QT.Qt.ControlModifier:
            #~ z = 10 if ev.delta()>0 else 1/10.
        #~ else:
            #~ z = 1.3 if ev.delta()>0 else 1/1.3
        #~ self.gain_zoom.emit(z)
        #~ ev.accept()
    def raiseContextMenu(self, ev):
        #for some reasons enableMenu=False is not taken (bug ????)
        pass


class FeatureTimeViewer(WidgetBase):

    _params = [
                    {'name': 'metric', 'type': 'list', 'values' : ['extremum_amplitude', 'geometry0', 'geometry1']},
                    {'name': 'alpha', 'type': 'float', 'value' : 0.5, 'limits':(0, 1.), 'step':0.05 },
                    {'name': 'spread', 'type': 'float', 'value' : 2.5, 'limits':(0, 50.), 'step':0.1 },
        ]
    
    
    def __init__(self, controller=None, parent=None):
        WidgetBase.__init__(self, parent=parent, controller=controller)

        self.layout = QT.QVBoxLayout()
        self.setLayout(self.layout)
        
        self.combo_seg = QT.QComboBox()
        self.layout.addWidget(self.combo_seg)
        self.combo_seg.addItems([ 'Segment {}'.format(seg_num) for seg_num in range(self.controller.dataio.nb_segment) ])
        self._seg_pos = 0
        self.seg_num = self._seg_pos
        self.combo_seg.currentIndexChanged.connect(self.refresh)
        
        
        self.graphicsview = pg.GraphicsView()
        self.layout.addWidget(self.graphicsview)
        
        self.create_settings()
        
        self.initialize_plot()
        self.similarity = None
        
        self.on_params_changed()#this do refresh
    
    
    def on_params_changed(self, ): #params, changes
        self.refresh()
        
    
    
    def initialize_plot(self):
        self.viewBox = MyViewBox()
        self.viewBox.doubleclicked.connect(self.open_settings)
        #~ self.viewBox.gain_zoom.connect(self.gain_zoom)
        self.viewBox.disableAutoRange()
        
        self.plot = pg.PlotItem(viewBox=self.viewBox)
        self.graphicsview.setCentralItem(self.plot)
        self.plot.hideButtons() 


    def refresh(self):
        
        self.plot.clear()
        if self.controller.some_peaks_index is None:
            return
        
        cluster_visible = self.controller.cluster_visible
        visibles = [c for c, v in cluster_visible.items() if v ]

        seg_index =  self.combo_seg.currentIndex()
        

        if self.controller.spikes is None:
            return
        
        some_spikes = self.controller.spikes[self.controller.some_peaks_index]
        selected = some_spikes['segment'] == seg_index
        if len(selected)==0:
            return
        spikes = some_spikes[selected]
        all_times = spikes['index'].astype('float64') / self.controller.dataio.sample_rate
        
        #~ selected = self.controller.spike_segment[self.controller.some_peaks_index]==seg_index
        #~ all_index = self.controller.spike_index[self.controller.some_peaks_index][selected]
        #~ all_times = all_index.astype('float64')/self.controller.dataio.sample_rate
        #~ all_labels = self.controller.spike_label[self.controller.some_peaks_index][selected]
        
        #TODO if None
        # this is a hack to speedup some_waveforms[selected]
        # because boolean selection is slow here ???
        #~ ind_selected, = np.nonzero(selected)
        #~ selected_slice = slice(np.min(ind_selected), np.max(ind_selected)+1)
        
        if self.params['metric'] == 'extremum_amplitude':
            #~ extremum_amplitude = spikes['extremum_amplitude']
            values = spikes['extremum_amplitude']
            
        elif self.params['metric'] == 'geometry0':
            geometry0 = self.controller.geometry[:, 0]
            values = geometry0[spikes['channel']]
            values += np.random.randn(values.size) * self.params['spread']
        elif self.params['metric'] == 'geometry1':
            geometry0 = self.controller.geometry[:, 1]
            values = geometry0[spikes['channel']]
            values += np.random.randn(values.size) * self.params['spread']
            
        else:
            raise NotImplementedError


        for k in visibles:
            #~ self.controller.some
            keep = spikes['cluster_label'] == k
            
            x = all_times[keep]
            y = values[keep]
            
            color = QT.QColor(self.controller.qcolors.get(k, QT.QColor( 'white')))
            color.setAlpha(int(self.params['alpha']*255))
            curve = pg.ScatterPlotItem(x=x, y=y, pen=pg.mkPen(color, width=1), brush=color)
            
            self.plot.addItem(curve)
            #~ self.curves.append(curve)
        
        
        self.plot.setXRange(0, all_times[-1])
        #~ if self.params['metric'] == 'extremum_amplitude':
            #~ self.plot.setYRange(-30, 30)
        #~ else:
        self.plot.setYRange(np.min(values), np.max(values))

    def on_spike_selection_changed(self):
        pass

    def on_spike_label_changed(self):
        self.refresh()
        
    def on_colors_changed(self):
        self.refresh()
    
    def on_cluster_visibility_changed(self):
        self.refresh()
    
    def on_cluster_tag_changed(self):
        pass






import shutil

from tridesclous.dataio import DataIO

from tridesclous.catalogueconstructor import CatalogueConstructor
from tridesclous import Peeler
from tridesclous.tests.testingtools import setup_catalogue
from tridesclous.tests.testingtools import ON_CI_CLOUD

from tridesclous.report import summary_catalogue_clusters, summary_noise, summary_after_peeler_clusters, generate_report
from tridesclous.autoparams import get_auto_params_for_peelers


import matplotlib.pyplot as plt


from tridesclous.tests.testingtools import ON_CI_CLOUD
import pytest


def setup_module():
    setup_catalogue('test_report', dataset_name='striatum_rat')
    
    dataio = DataIO(dirname='test_report')
    catalogue = dataio.load_catalogue(chan_grp=0)
    params = get_auto_params_for_peelers(dataio, chan_grp=0)
    peeler = Peeler(dataio)
    peeler.change_params(catalogue=catalogue, **params)
    peeler.run(progressbar=False)

def teardown_module():
    shutil.rmtree('test_report')

@pytest.mark.skipif(ON_CI_CLOUD, reason='ON_CI_CLOUD')
def test_summary_catalogue_clusters():
    dataio = DataIO(dirname='test_report')

    #~ summary_catalogue_clusters(dataio, chan_grp=0)
    summary_catalogue_clusters(dataio, chan_grp=0, labels= [0])

@pytest.mark.skipif(ON_CI_CLOUD, reason='ON_CI_CLOUD')
def test_summary_noise():
    dataio = DataIO(dirname='test_report')
 
    
    summary_noise(dataio, chan_grp=0)

@pytest.mark.skipif(ON_CI_CLOUD, reason='ON_CI_CLOUD')
def test_summary_after_peeler_clusters():
    dataio = DataIO(dirname='test_report')
    summary_after_peeler_clusters(dataio, chan_grp=0,  labels= [0])
    
    summary_after_peeler_clusters(dataio, chan_grp=0,  labels= [0], neighborhood_radius=200)


@pytest.mark.skipif(ON_CI_CLOUD, reason='ON_CI_CLOUD')
def test_generate_report():
    dataio = DataIO(dirname='test_report')
    
    generate_report(dataio)
    

if __name__ == '__main__':
    
    setup_module()
    
    test_summary_catalogue_clusters()
    
    test_summary_noise()
    
    test_summary_after_peeler_clusters()
    
    #~ plt.show()
    
    test_generate_report()
    
    #~ plt.show()




from .myqt import QT
import pyqtgraph as pg

import numpy as np
import pandas as pd

from .base import WidgetBase



class MyViewBox(pg.ViewBox):
    doubleclicked = QT.pyqtSignal()
    def mouseDoubleClickEvent(self, ev):
        self.doubleclicked.emit()
        ev.accept()
    def raiseContextMenu(self, ev):
        #for some reasons enableMenu=False is not taken (bug ????)
        pass



class ISIViewer(WidgetBase):
    _params = [
                      {'name': 'bin_min', 'type': 'float', 'value' : 0. },
                      {'name': 'bin_max', 'type': 'float', 'value' : 100. },
                      {'name': 'bin_size', 'type': 'float', 'value' : 1.0 },
        ]
    def __init__(self, controller=None, parent=None):
        WidgetBase.__init__(self, parent=parent, controller=controller)
        
        self.layout = QT.QVBoxLayout()
        self.setLayout(self.layout)
        
        #~ h = QT.QHBoxLayout()
        #~ self.layout.addLayout(h)
        #~ h.addWidget(QT.QLabel('<b>Similarity</b>') )

        #~ but = QT.QPushButton('settings')
        #~ but.clicked.connect(self.open_settings)
        #~ h.addWidget(but)
        
        
        self.graphicsview = pg.GraphicsView()
        self.layout.addWidget(self.graphicsview)
        
        self.initialize_plot()
        
        #~ self.on_params_changed()#this do refresh    


    def initialize_plot(self):
        self.viewBox = MyViewBox()
        self.viewBox.doubleclicked.connect(self.open_settings)
        #~ self.viewBox.disableAutoRange()
        
        self.plot = pg.PlotItem(viewBox=self.viewBox)
        self.graphicsview.setCentralItem(self.plot)
        self.plot.hideButtons()
        
        #ISI are computed on demand
        self.all_isi = {}
    
    def _compute_isi(self, k):
        spikes = self.controller.spikes
        
        isi = []
        for seg_num in range(self.controller.dataio.nb_segment):
            sel = (spikes['segment'] == seg_num) & (spikes['cluster_label'] == k)
            isi.append(np.diff(spikes[sel]['index'])/self.controller.dataio.sample_rate)
        isi = np.concatenate(isi)
        self.all_isi[k] = isi * 1000. #ms

    def refresh(self):
        self.plot.clear()
        
        n = 0
        for k in self.controller.positive_cluster_labels:
            if not self.controller.cluster_visible[k]:
                continue
            
            if k not in self.all_isi:
                self._compute_isi(k)
            
            isi = self.all_isi[k]
            if len(isi) ==0:
                return
            
            bins = np.arange(self.params['bin_min'], self.params['bin_max'], self.params['bin_size'])
            count, bins = np.histogram(isi, bins=bins)
            
            qcolor = self.controller.qcolors[k]
            curve = pg.PlotCurveItem(bins[:-1], count, pen=pg.mkPen(qcolor, width=3))
            self.plot.addItem(curve)





import time

from .peeler_tools import *
from .peeler_tools import _dtype_spike
from .tools import make_color_dict

from .signalpreprocessor import signalpreprocessor_engines
#~ from .peakdetector import get_peak_detector_class


from .cltools import HAVE_PYOPENCL, OpenCL_Helper
if HAVE_PYOPENCL:
    import pyopencl
    mf = pyopencl.mem_flags

from . import pythran_tools
if hasattr(pythran_tools, '__pythran__'):
    HAVE_PYTHRAN = True
else:
    HAVE_PYTHRAN = False

try:
    import numba
    HAVE_NUMBA = True
    from .numba_tools import numba_loop_sparse_dist
except ImportError:
    HAVE_NUMBA = False





class PeelerEngineBase(OpenCL_Helper):
    def change_params(self, catalogue=None,
                                        chunksize=1024, 
                                        internal_dtype='float32', 
                                        maximum_jitter_shift = 4,
                                        
                                        save_bad_label=False,
                                        
                                        cl_platform_index=None,
                                        cl_device_index=None,
                                        ):
        """
        Set parameters for the Peeler.
        
        
        Parameters
        ----------
        catalogue: the catalogue (a dict)
            The catalogue made by CatalogueConstructor.
        chunksize: int (1024 by default)
            the size of chunk for processing.
        internal_dtype: 'float32' or 'float64'
            dtype of internal processing. float32 is OK. float64 is totally useless.
        maximum_jitter_shift
            Maximum allowed shift alignement between peak index and its template.
            
        """
        assert catalogue is not None
        self.catalogue = catalogue
        self.chunksize = chunksize
        self.internal_dtype= internal_dtype
        self.maximum_jitter_shift = maximum_jitter_shift
        self.save_bad_label = save_bad_label
        
        self.inter_sample_oversampling = self.catalogue['inter_sample_oversampling']
        
        self.cl_platform_index=None
        self.cl_device_index=None
        
        if self.catalogue['mode'] == 'sparse':
            self.use_sparse_template = True
        elif self.catalogue['mode'] == 'dense':
            self.use_sparse_template = False
        
        self.colors = make_color_dict(self.catalogue['clusters'])
        
        
        # DEBUG
        #~ print('DEBUG peakdetector all neigboor')
        #~ self.catalogue['peak_detector_params']['adjacency_radius_um'] = 0.
        #~ self.catalogue['peak_detector_params']['peak_span_ms'] = 0.
        # / DEBUG
        
        # precompute some value for jitter estimation
        if self.inter_sample_oversampling:
            n = self.catalogue['cluster_labels'].size
            self.catalogue['wf1_norm2'] = np.zeros(n)
            self.catalogue['wf2_norm2'] = np.zeros(n)
            self.catalogue['wf1_dot_wf2'] = np.zeros(n)
            for i, k in enumerate(self.catalogue['cluster_labels']):
                chan = self.catalogue['extremum_channel'][i]
                wf0 = self.catalogue['centers0'][i,: , chan]
                wf1 = self.catalogue['centers1'][i,: , chan]
                wf2 = self.catalogue['centers2'][i,: , chan]

                self.catalogue['wf1_norm2'][i] = wf1.dot(wf1)
                self.catalogue['wf2_norm2'][i] = wf2.dot(wf2)
                self.catalogue['wf1_dot_wf2'][i] = wf1.dot(wf2)
        
        
        #~ print('self.use_sparse_template', self.use_sparse_template)

        # make kernels
        #~ centers = self.catalogue['centers0']
        #~ self.nb_cluster = centers.shape[0]
        #~ self.peak_width = centers.shape[1]
        #~ self.nb_channel = centers.shape[2]
        
        
        self.threshold = self.catalogue['peak_detector_params']['relative_threshold']
        
        #~ print(centers.shape)
        #~ abs_centers = np.abs(centers)

        self.sparse_mask_level1 = self.catalogue['sparse_mask_level1']
        self.sparse_mask_level2 = self.catalogue['sparse_mask_level2']
        self.sparse_mask_level3 = self.catalogue['sparse_mask_level3']
        

        
        # weight of template per channel masked with level3
        #~ self.weight_per_template = np.zeros((self.nb_cluster, self.nb_channel), dtype='float32')
        #~ self.weight_per_template_dict = {}
        #~ centers = self.catalogue['centers0']
        #~ for i, k in enumerate(self.catalogue['cluster_labels']):
            #~ mask = self.sparse_mask_level2[i, :]
            #~ wf = centers[i, :, :][:, mask]
            #~ w = np.sum(wf**2, axis=0)
            #~ self.weight_per_template[i, mask] = w
            #~ self.weight_per_template_dict[i] = w
        
        
        self.projections = self.catalogue['projections']
        self.boundaries = self.catalogue['boundaries']
        

        #~ import matplotlib.pyplot as plt
        #~ for i in range(centers.shape[0]):
            #~ fig, ax = plt.subplots()
            #~ center = centers[i,:,:].copy()
            #~ center_sparse = center.copy()
            #~ center_sparse[:, ~mask[i, :]] = 0.
            #~ ax.plot(center.T.flatten(), color='g')
            #~ ax.plot(center_sparse.T.flatten(), color='r', ls='--')
            #~ ax.axhline(sparse_threshold_mad)
            #~ ax.axhline(-sparse_threshold_mad)
            #~ plt.show()
    
    def initialize(self, sample_rate=None, nb_channel=None, source_dtype=None, geometry=None, already_processed=False, processor_engine=None):
        self.nb_channel = nb_channel
        self.sample_rate = sample_rate
        self.source_dtype = source_dtype
        self.geometry = geometry
        self.already_processed = already_processed # this is globally set but can be change segment per segment
        

        if not self.already_processed:
            # signal processor class
            p = dict(self.catalogue['signal_preprocessor_params'])
            self.signalpreprocessor_engine = p.pop('engine')
            if processor_engine is not None:
                # can be force to opencl by geometrial_opencl
                self.signalpreprocessor_engine = processor_engine
            
            SignalPreprocessor_class = signalpreprocessor_engines[self.signalpreprocessor_engine]
            self.signalpreprocessor = SignalPreprocessor_class(sample_rate, nb_channel, self.chunksize, source_dtype)
            p['normalize'] = True
            p['signals_medians'] = self.catalogue['signals_medians']
            p['signals_mads'] = self.catalogue['signals_mads']
            
            if hasattr(self, 'ctx') and self.ctx is not None and self.signalpreprocessor_engine == 'opencl':
                # use local ctx and queue if exists for processor
                #~ print('yep', self.ctx)
                p['cl_platform_index'] = None
                p['cl_device_index'] = None
                p['ctx'] = self.ctx
                p['queue'] = self.queue
            
            self.signalpreprocessor.change_params(**p)
            self.internal_dtype = self.signalpreprocessor.output_dtype
            
            assert self.chunksize>self.signalpreprocessor.pad_width, 'pad_width ({}) is greater than chunksize ({})!'.format(self.signalpreprocessor.pad_width, self.chunksize)
            
        else:
            # no need
            self.signalpreprocessor = None
            self.internal_dtype = source_dtype

        # peak detector class
        self.peak_sign = self.catalogue['peak_detector_params']['peak_sign']
        self.relative_threshold = self.catalogue['peak_detector_params']['relative_threshold']
        peak_span_ms = self.catalogue['peak_detector_params']['peak_span_ms']
        self.n_span = int(sample_rate * peak_span_ms / 1000.)//2
        self.n_span = max(1, self.n_span)
        #~ self.peak_width = self.catalogue['peak_width']
        
        centers = self.catalogue['centers0']
        self.nb_cluster = centers.shape[0]
        #~ self.peak_width = centers.shape[1]
        self.nb_channel = centers.shape[2]
        
        self.n_left = self.catalogue['n_left']
        self.n_right = self.catalogue['n_right']
        self.n_left_long = self.catalogue['n_left_long']
        self.n_right_long = self.catalogue['n_right_long']
        self.peak_width = self.n_right - self.n_left
        self.peak_width_long = self.n_right_long - self.n_left_long
        
        self.extra_size = self.peak_width_long + self.maximum_jitter_shift*2 + self.n_span*2 + 1
        self.fifo_size = self.chunksize + self.extra_size
        
        assert self.chunksize > (self.extra_size+1), 'chunksize is too small because of n_size'
        
        alien_thresh = self.catalogue['clean_peaks_params']['alien_value_threshold']
        if alien_thresh is not None:
            if (alien_thresh < 0.) or np.isnan(alien_thresh):
                # -1 is equivalent to None
                alien_thresh = None
        self.alien_value_threshold = alien_thresh
        
        self.total_spike = 0
        
        self.near_border_good_spikes = []
        
        self.fifo_residuals = np.zeros((self.fifo_size, nb_channel), dtype=self.internal_dtype)
    
    def initialize_before_each_segment(self, already_processed=False):
        self.total_spike = 0
        self.near_border_good_spikes = []
        self.fifo_residuals = np.zeros((self.fifo_size, self.nb_channel), dtype=self.internal_dtype)
        
        if self.signalpreprocessor is not None:
            self.signalpreprocessor.initialize_stream()
        
        self.already_processed = already_processed

    def get_remaining_spikes(self):
        if len(self.near_border_good_spikes)>0:
            # deal with extra remaining spikes
            extra_spikes = self.near_border_good_spikes[0]
            extra_spikes = extra_spikes.take(np.argsort(extra_spikes['index']))
            self.total_spike += extra_spikes.size
            return extra_spikes




class PeelerEngineGeneric(PeelerEngineBase):
    # common base for PeelerEngineGeometrical and PeelerEngineClassic 
    # andPeelerEngineGeometricalCl

    def process_one_chunk(self,  pos, sigs_chunk):
        #~ if 16000 <pos<16400:
            #~ self._plot_debug = True
        #~ else:
            #~ self._plot_debug = False
        self._plot_debug = False
        #~ self._plot_debug = True
        
        #~ if pos >= 732 * self.chunksize:
            #~ self._plot_debug = True
            
            
            
        if self._plot_debug:
        #~ if True:
            print('*'*10)
            print('process_one_chunk', pos)
        
        abs_head_index, preprocessed_chunk = self.apply_processor( pos, sigs_chunk)
        
        # relation between inside chunk index and abs index
        to_local_shift = abs_head_index - self.fifo_size
        
        
        #~ t1 = time.perf_counter()
        self.detect_local_peaks_before_peeling_loop()
        #~ t2 = time.perf_counter()
        #~ print()
        #~ print('  detect_local_peaks_before_peeling_loop', (t2-t1)*1000)
        
        #~ self._debug_nb_accept_tempate = 0
        
        good_spikes = []
        

        if self._plot_debug:
            self._plot_before_peeling_loop()

        n_loop = 0
        t3 = time.perf_counter()
        
        while True: # main loop
            if self._plot_debug:
            #~ if True:
                print('** peeler level +1 **')
            nb_good_spike = 0
            
            
            # loop : one more peeler level
            while True: 
                #~ print()
                #~ t1 = time.perf_counter()
                spike = self.classify_and_align_next_spike()
                #~ t2 = time.perf_counter()
                #~ print('  classify_and_align_next_spike', (t2-t1)*1000, spike)
                #~ if spike.cluster_label <0:
                    #~ print('   spike.label', spike.cluster_label, 'spike.index', spike.index)

                if spike.cluster_label == LABEL_NO_MORE_PEAK:
                    break
                
                if (spike.cluster_label >=0):
                    #~ good_spikes.append(np.array([spike], dtype=_dtype_spike))
                    good_spikes.append(spike)
                    nb_good_spike+=1
                n_loop +=1 
            
            #~ if self._plot_debug:
                #~ print('***end inner loop', n_loop,'nb_good_spike', nb_good_spike)
            
            if nb_good_spike == 0:
                break
            else:
                #~ t1 = time.perf_counter()
                self.reset_to_not_tested(good_spikes[-nb_good_spike:])
                #~ t2 = time.perf_counter()
                #~ print('  reset_to_not_tested', (t2-t1)*1000)
            
            if self._plot_debug:
                self._plot_after_inner_peeling_loop()
        

        #~ t4 = time.perf_counter()
        #~ print('mainloop classify_and_align ', len(good_spikes), ' spike', (t4-t3)*1000, 'ms', 'n_loop', n_loop)
        
        if self._plot_debug:
        #~ if True:
            t4 = time.perf_counter()
            print('mainloop classify_and_align ', len(good_spikes), ' spike', (t4-t3)*1000, 'ms', 'n_loop', n_loop)
            self._plot_after_peeling_loop(good_spikes)
        
        #~ print(self._debug_nb_accept_tempate)
        #~ if  len(good_spikes)>0:
            #~ print('nb_good_spike', len(good_spikes), 'n_loop', n_loop, 'per spike', (t4-t3)*1000/len(good_spikes))
        
        if self.save_bad_label:
            bad_spikes = self.get_no_label_peaks()
            bad_spikes['index'] += to_local_shift
        
        
        if len(good_spikes)>0:
            # TODO remove from peak the very begining of the signal because of border filtering effects
            
            good_spikes = np.array(good_spikes, dtype=_dtype_spike)
            good_spikes['index'] += to_local_shift
            near_border = (good_spikes['index'] - to_local_shift)>=(self.chunksize+self.n_span)
            near_border_good_spikes = good_spikes[near_border].copy()
            good_spikes = good_spikes[~near_border]
            
            if self.save_bad_label:
                all_spikes = np.concatenate([good_spikes] + [bad_spikes] + self.near_border_good_spikes)
            else:
                all_spikes = np.concatenate([good_spikes] + self.near_border_good_spikes)
            self.near_border_good_spikes = [near_border_good_spikes] # for next chunk
        else:
            if self.save_bad_label:
                all_spikes = np.concatenate([bad_spikes] + self.near_border_good_spikes)
            elif len(self.near_border_good_spikes) > 0:
                all_spikes = np.concatenate(self.near_border_good_spikes)
            else:
                all_spikes = np.array([], dtype=_dtype_spike)
            self.near_border_good_spikes = []
        
        # all_spikes = all_spikes[np.argsort(all_spikes['index'])]
        all_spikes = all_spikes.take(np.argsort(all_spikes['index']))
        self.total_spike += all_spikes.size
        
        #~ print(good_spikes.size, all_spikes.size)
        return abs_head_index, preprocessed_chunk, self.total_spike, all_spikes

    def apply_processor(self, pos, sigs_chunk):
        if self.already_processed:
            abs_head_index, preprocessed_chunk =  pos, sigs_chunk
        else:
            abs_head_index, preprocessed_chunk = self.signalpreprocessor.process_buffer_stream(pos, sigs_chunk)
        
        #shift residuals buffer and put the new one on right side
        fifo_roll_size = self.fifo_size-preprocessed_chunk.shape[0]
        if fifo_roll_size>0 and fifo_roll_size!=self.fifo_size:
            self.fifo_residuals[:fifo_roll_size,:] = self.fifo_residuals[-fifo_roll_size:,:]
            self.fifo_residuals[fifo_roll_size:,:] = preprocessed_chunk
        
        return abs_head_index, preprocessed_chunk 


    def classify_and_align_next_spike(self):
        
        
        if self._plot_debug:
            print('classify_and_align_next_spike')
        
        proposed_peak_ind, peak_chan = self.select_next_peak()
        if self._plot_debug:        
            print('peak ', 'index',  proposed_peak_ind, 'chan', peak_chan)
            
        if proposed_peak_ind == LABEL_NO_MORE_PEAK:
            return Spike(0, LABEL_NO_MORE_PEAK, 0)
        
        # left_ind is the waveform left border
        left_ind = proposed_peak_ind + self.n_left
        left_ind_long = proposed_peak_ind + self.n_left_long

        if left_ind_long+self.peak_width_long+self.maximum_jitter_shift+1>=self.fifo_size:
        #~ if left_ind+self.peak_width >=self.fifo_size:
            # TODO : remove this because maybe unecessry
            # too near right limits no label
            label = LABEL_RIGHT_LIMIT
            jitter = 0
            if self._plot_debug:
                print('LABEL_RIGHT_LIMIT', proposed_peak_ind, peak_chan)
        elif left_ind_long<=self.maximum_jitter_shift:
        #~ elif left_ind<0:
            # TODO : remove this because maybe unecessry
            # too near left limits no label
            #~ print('     LABEL_LEFT_LIMIT', left_ind)
            label = LABEL_LEFT_LIMIT
            jitter = 0
            if self._plot_debug:
                print('LABEL_LEFT_LIMIT', proposed_peak_ind, peak_chan)
        elif self.catalogue['centers0'].shape[0]==0:
            # empty catalogue
            label  = LABEL_UNCLASSIFIED
            jitter = 0
            #~ if self._plot_debug:
                #~ print('LABEL_UNCLASSIFIED', proposed_peak_ind, peak_chan)
        else:
            #~ t1 = time.perf_counter()
            waveform = self.fifo_residuals[left_ind:left_ind+self.peak_width,:]
            #~ t2 = time.perf_counter()
            #~ print('    get_waveform', (t2-t1)*1000)
            
            #~ if self.alien_value_threshold is not None and self.alien_value_threshold > 0 and \
            if self.alien_value_threshold is not None and \
                    np.any((waveform>self.alien_value_threshold) | (waveform<-self.alien_value_threshold)) :
                label  = LABEL_ALIEN
                jitter = 0
                #~ if self._plot_debug:
                    #~ print('LABEL_ALIEN', proposed_peak_ind, peak_chan)

            else:
                
                #~ t1 = time.perf_counter()
                #TODO try usewaveform to avoid new buffer ????
                
                cluster_idx, shift, best_template_info = self.get_best_template(left_ind, peak_chan)
                
                
                if shift is not None:
                    left_ind += shift
                    left_ind_long +=shift
                
                # cluster_idx can be <0 when no possible cluster with geometrical
                
                #~ t2 = time.perf_counter()
                #~ print('    get_best_template', (t2-t1)*1000)
                


                
                if cluster_idx is None or cluster_idx<0:
                    #~ print('LABEL_UNCLASSIFIED cluster_idx<0')
                    label  = LABEL_UNCLASSIFIED
                    jitter = 0
                    if self._plot_debug:
                    #~ if True:
                        self._plot_label_unclassified(left_ind, peak_chan, cluster_idx, jitter)
                    
                else:
                    label = None
                    #~ t1 = time.perf_counter()
                    #~ print('left_ind', left_ind, 'proposed_peak_ind', proposed_peak_ind)
                    if self.inter_sample_oversampling:
                        raise NotImplementedError # TODO propagate center_long to this section and peeler
                        jitter = self.estimate_jitter(left_ind, cluster_idx)
                        shift = -int(np.round(jitter))
                        
                        if (np.abs(jitter) > 0.5) and \
                                        (left_ind_long+shift+self.peak_width_long<self.fifo_size) and\
                                        ((left_ind_long + shift) >= 0):
                            # try better jitter
                            new_jitter = self.estimate_jitter(left_ind + shift, cluster_idx)
                            if np.abs(new_jitter)<np.abs(jitter):
                                jitter = new_jitter
                                left_ind += shift
                                left_ind_long += shift
                                shift = -int(np.round(new_jitter))
                        
                        # security to not be outside the fifo
                        if np.abs(shift) > self.maximum_jitter_shift:
                            label = LABEL_MAXIMUM_SHIFT
                            if self._plot_debug:
                                print('LABEL_MAXIMUM_SHIFT', proposed_peak_ind, peak_chan)
                        elif (left_ind_long+shift+self.peak_width_long)>=self.fifo_size:
                            # normally this should be resolve in the next chunk
                            label = LABEL_RIGHT_LIMIT
                            if self._plot_debug:
                                print('LABEL_RIGHT_LIMIT 2', proposed_peak_ind, peak_chan)
                        elif (left_ind_long + shift) < 0:
                            label = LABEL_LEFT_LIMIT
                            if self._plot_debug:
                                print('LABEL_LEFT_LIMIT 2', proposed_peak_ind, peak_chan)
                    else:
                        jitter = None
                    
                    if label is None:
                        #~ t1 = time.perf_counter()
                        ok = self.accept_tempate(left_ind, cluster_idx, jitter, best_template_info)
                        if ok:
                            label = self.catalogue['cluster_labels'][cluster_idx]
                        else:
                            label  = LABEL_UNCLASSIFIED
                            jitter = 0
                            if self._plot_debug:
                                self._plot_label_unclassified(left_ind, peak_chan, cluster_idx, jitter)
                        #~ t2 = time.perf_counter()
                        #~ print('    accept_tempate', (t2-t1)*1000)
        
        # second security check for borders
        if label>=0 and jitter is not None:
            left_ind_check = left_ind_long - np.round(jitter).astype('int64')
            if left_ind_check<0:
                label = LABEL_LEFT_LIMIT
                if self._plot_debug:
                    print('!!!!!!!ici LABEL_LEFT_LIMIT', label)

            elif (left_ind_check+self.peak_width_long) >=self.fifo_size:
                label = LABEL_RIGHT_LIMIT
                if self._plot_debug:
                    print('!!!!!!!ici LABEL_RIGHT_LIMIT', label)
                
        
        #~ if self._plot_debug:
            #~ import matplotlib.pyplot as plt
            #~ if label in (LABEL_LEFT_LIMIT, LABEL_RIGHT_LIMIT, LABEL_UNCLASSIFIED):
                #~ fig, ax = plt.subplots()
                #~ waveform = self.fifo_residuals[left_ind:left_ind+self.peak_width,:]
                #~ ax.plot(waveform.T.flatten())
                
                #~ if label == LABEL_LEFT_LIMIT:
                    #~ ax.set_title('LABEL_LEFT_LIMIT')
                #~ if label == LABEL_RIGHT_LIMIT:
                    #~ ax.set_title('LABEL_RIGHT_LIMIT')
                #~ if label == LABEL_UNCLASSIFIED:
                    #~ ax.set_title('LABEL_UNCLASSIFIED')
                    


                #~ max_chan_ind = np.argmax(np.abs(waveform[-self.n_left, :]))
                #~ fig, ax = plt.subplots()
                #~ ax.plot(self.fifo_residuals[:, max_chan_ind])
                
                #~ ax.scatter([left_ind-self.n_left], [self.fifo_residuals[left_ind-self.n_left, max_chan_ind]], color='r')
                
                
                #~ plt.show()
        
        
        if label < 0:
            # set peak tested to not test it again
            #~ self.mask_not_already_tested[proposed_peak_ind - self.n_span] = False
            peak_ind = proposed_peak_ind
            #~ jitter = peak_chan
            #~ t1 = time.perf_counter()
            self.set_already_tested(peak_ind, peak_chan)
            #~ t2 = time.perf_counter()
            #~ print('    set_already_tested', (t2-t1)*1000)


        #~ self.update_peak_mask(peak_ind, label)
        #~ t2 = time.perf_counter()
        #~ print('    update_peak_mask', (t2-t1)*1000)
        else:
            if jitter is not None:
                # ensure jitter in range [-0.5, 0.5]
                shift = -int(np.round(jitter))
                if shift !=0:
                    jitter = jitter + shift
                    left_ind = left_ind + shift
                    left_ind_long = left_ind_long + shift
            
            peak_ind = left_ind - self.n_left
            
            
            # remove from residulals
            #~ t1 = time.perf_counter()
            self.on_accepted_spike(peak_ind, cluster_idx, jitter)
            #~ t2 = time.perf_counter()
            #~ print('    on_accepted_spike', (t2-t1)*1000)
            
        
        if self._plot_debug:
            print('Spike', peak_ind, label, jitter)
        
        return Spike(peak_ind, label, jitter)


    def estimate_jitter(self, left_ind, cluster_idx):
        
        chan_max = self.catalogue['extremum_channel'][cluster_idx]
        
        wf0 = self.catalogue['centers0'][cluster_idx,: , chan_max]
        wf1 = self.catalogue['centers1'][cluster_idx,: , chan_max]
        wf2 = self.catalogue['centers2'][cluster_idx,: , chan_max]

        wf = self.fifo_residuals[left_ind:left_ind+self.peak_width,chan_max]
        
        
        #it is  precompute that at init for speedup
        wf1_norm2= self.catalogue['wf1_norm2'][cluster_idx]
        wf2_norm2 = self.catalogue['wf2_norm2'][cluster_idx]
        wf1_dot_wf2 = self.catalogue['wf1_dot_wf2'][cluster_idx]
        
        h = wf - wf0
        h0_norm2 = h.dot(h)
        h_dot_wf1 = h.dot(wf1)
        jitter0 = h_dot_wf1/wf1_norm2
        h1_norm2 = np.sum((h-jitter0*wf1)**2)
        #~ print(h0_norm2, h1_norm2)
        #~ print(h0_norm2 > h1_norm2)
        
        if h0_norm2 > h1_norm2:
            #order 1 is better than order 0
            h_dot_wf2 = np.dot(h,wf2)
            rss_first = -2*h_dot_wf1 + 2*jitter0*(wf1_norm2 - h_dot_wf2) + 3*jitter0**2*wf1_dot_wf2 + jitter0**3*wf2_norm2
            rss_second = 2*(wf1_norm2 - h_dot_wf2) + 6*jitter0*wf1_dot_wf2 + 3*jitter0**2*wf2_norm2
            jitter1 = jitter0 - rss_first/rss_second
            #~ h2_norm2 = np.sum((h-jitter1*wf1-jitter1**2/2*wf2)**2)
            #~ if h1_norm2 <= h2_norm2:
                #when order 2 is worse than order 1
                #~ jitter1 = jitter0
        else:
            jitter1 = 0.
        
        #~ print(jitter1)
        return jitter1

